from collections import Counter

import pytest
from numpy.testing import assert_equal

from brian2 import *
from brian2.core.functions import timestep
from brian2.devices.device import reinit_and_delete
from brian2.equations.refractory import add_refractoriness
from brian2.tests.utils import assert_allclose, exc_isinstance
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_add_refractoriness():
    eqs = Equations(
        """
        dv/dt = -x*v/second : volt (unless refractory)
        dw/dt = -w/second : amp
        x : 1
        """
    )
    # only make sure it does not throw an error
    eqs = add_refractoriness(eqs)
    # Check that the parameters were added
    assert "not_refractory" in eqs
    assert "lastspike" in eqs


@pytest.mark.codegen_independent
def test_missing_refractory_warning():
    # Forgotten refractory argument
    with catch_logs() as l:
        group = NeuronGroup(
            1,
            "dv/dt = -v / (10*ms) : 1 (unless refractory)",
            threshold="v > 1",
            reset="v = 0",
        )
    assert len(l) == 1
    assert l[0][0] == "WARNING" and l[0][1].endswith("no_refractory")


@pytest.mark.standalone_compatible
def test_refractoriness_basic():
    G = NeuronGroup(
        1,
        """
        dv/dt = 99.999*Hz : 1 (unless refractory)
        dw/dt = 99.999*Hz : 1
        """,
        threshold="v>1",
        reset="v=0;w=0",
        refractory=5 * ms,
    )
    # It should take 10ms to reach the threshold, then v should stay at 0
    # for 5ms, while w continues to increase
    mon = StateMonitor(G, ["v", "w"], record=True, when="end")
    run(20 * ms)
    # No difference before the spike
    assert_allclose(
        mon[0].v[: timestep(10 * ms, defaultclock.dt)],
        mon[0].w[: timestep(10 * ms, defaultclock.dt)],
    )
    # v is not updated during refractoriness
    in_refractoriness = mon[0].v[
        timestep(10 * ms, defaultclock.dt) : timestep(15 * ms, defaultclock.dt)
    ]
    assert_equal(in_refractoriness, np.zeros_like(in_refractoriness))
    # w should evolve as before
    assert_allclose(
        mon[0].w[: timestep(5 * ms, defaultclock.dt)],
        mon[0].w[
            timestep(10 * ms, defaultclock.dt)
            + 1 : timestep(15 * ms, defaultclock.dt)
            + 1
        ],
    )
    assert np.all(
        mon[0].w[
            timestep(10 * ms, defaultclock.dt)
            + 1 : timestep(15 * ms, defaultclock.dt)
            + 1
        ]
        > 0
    )
    # After refractoriness, v should increase again
    assert np.all(
        mon[0].v[
            timestep(15 * ms, defaultclock.dt) : timestep(20 * ms, defaultclock.dt)
        ]
        > 0
    )


@pytest.mark.standalone_compatible
@pytest.mark.parametrize(
    "ref_time",
    [
        "5*ms",
        "(t-lastspike + 1e-3*dt) < 5*ms",
        "time_since_spike + 1e-3*dt < 5*ms",
        "ref_subexpression",
        "(t-lastspike + 1e-3*dt) < ref",
        "ref",
        "ref_no_unit*ms",
    ],
)
def test_refractoriness_variables(ref_time):
    # Try a string evaluating to a quantity, and an explicit boolean
    # condition -- all should do the same thing
    G = NeuronGroup(
        1,
        """
        dv/dt = 99.999*Hz : 1 (unless refractory)
        dw/dt = 99.999*Hz : 1
        ref : second
        ref_no_unit : 1
        time_since_spike = (t - lastspike) +1e-3*dt : second
        ref_subexpression = (t - lastspike + 1e-3*dt) < ref : boolean
        """,
        threshold="v>1",
        reset="v=0;w=0",
        refractory=ref_time,
        dtype={
            "ref": defaultclock.variables["t"].dtype,
            "ref_no_unit": defaultclock.variables["t"].dtype,
            "lastspike": defaultclock.variables["t"].dtype,
            "time_since_spike": defaultclock.variables["t"].dtype,
        },
    )
    G.ref = 5 * ms
    G.ref_no_unit = 5
    # It should take 10ms to reach the threshold, then v should stay at 0
    # for 5ms, while w continues to increase
    mon = StateMonitor(G, ["v", "w"], record=True, when="end")
    run(20 * ms)
    try:
        # No difference before the spike
        assert_allclose(
            mon[0].v[: timestep(10 * ms, defaultclock.dt)],
            mon[0].w[: timestep(10 * ms, defaultclock.dt)],
        )
        # v is not updated during refractoriness
        in_refractoriness = mon[0].v[
            timestep(10 * ms, defaultclock.dt) : timestep(15 * ms, defaultclock.dt)
        ]
        assert_allclose(in_refractoriness, np.zeros_like(in_refractoriness))
        # w should evolve as before
        assert_allclose(
            mon[0].w[: timestep(5 * ms, defaultclock.dt)],
            mon[0].w[
                timestep(10 * ms, defaultclock.dt)
                + 1 : timestep(15 * ms, defaultclock.dt)
                + 1
            ],
        )
        assert np.all(
            mon[0].w[
                timestep(10 * ms, defaultclock.dt)
                + 1 : timestep(15 * ms, defaultclock.dt)
                + 1
            ]
            > 0
        )
        # After refractoriness, v should increase again
        assert np.all(
            mon[0].v[
                timestep(15 * ms, defaultclock.dt) : timestep(20 * ms, defaultclock.dt)
            ]
            > 0
        )
    except AssertionError as ex:
        raise
        raise AssertionError(
            f"Assertion failed when using {ref_time!r} as refractory argument:\n{ex}"
        )


@pytest.mark.standalone_compatible
def test_refractoriness_threshold_basic():
    G = NeuronGroup(
        1,
        """
        dv/dt = 199.99*Hz : 1
        """,
        threshold="v > 1",
        reset="v=0",
        refractory=10 * ms,
    )
    # The neuron should spike after 5ms but then not spike for the next
    # 10ms. The state variable should continue to integrate so there should
    # be a spike after 15ms
    spike_mon = SpikeMonitor(G)
    run(16 * ms)
    assert_allclose(spike_mon.t, [5, 15] * ms)


@pytest.mark.standalone_compatible
def test_refractoriness_repeated():
    # Create a group that spikes whenever it can
    group = NeuronGroup(1, "", threshold="True", refractory=10 * defaultclock.dt)
    spike_mon = SpikeMonitor(group)
    run(10000 * defaultclock.dt)
    assert spike_mon.t[0] == 0 * ms
    assert_allclose(np.diff(spike_mon.t), 10 * defaultclock.dt)


@pytest.mark.standalone_compatible
def test_refractoriness_repeated_legacy():
    if prefs.core.default_float_dtype == np.float32:
        pytest.skip(
            "Not testing legacy refractory mechanism with single precision floats."
        )
    # Switch on behaviour from versions <= 2.1.2
    prefs.legacy.refractory_timing = True
    # Create a group that spikes whenever it can
    group = NeuronGroup(1, "", threshold="True", refractory=10 * defaultclock.dt)
    spike_mon = SpikeMonitor(group)
    run(10000 * defaultclock.dt)
    assert spike_mon.t[0] == 0 * ms

    # Empirical values from running with earlier Brian versions
    assert_allclose(
        np.diff(spike_mon.t)[:10], [1.1, 1, 1.1, 1, 1.1, 1.1, 1.1, 1.1, 1, 1.1] * ms
    )
    steps = Counter(np.diff(np.int_(np.round(spike_mon.t / defaultclock.dt))))
    assert len(steps) == 2 and steps[10] == 899 and steps[11] == 91
    prefs.legacy.refractory_timing = False


@pytest.mark.standalone_compatible
@pytest.mark.parametrize(
    "ref_time",
    [
        10 * ms,
        "10*ms",
        "timestep(t-lastspike, dt) < timestep(10*ms, dt)",
        "timestep(t-lastspike, dt) < timestep(ref, dt)",
        "ref",
        "ref_no_unit*ms",
    ],
)
def test_refractoriness_threshold(ref_time):
    # Try a quantity, a string evaluating to a quantity, and an explicit boolean
    # condition -- all should do the same thing
    G = NeuronGroup(
        1,
        """
        dv/dt = 199.999*Hz : 1
        ref : second
        ref_no_unit : 1
        """,
        threshold="v > 1",
        reset="v=0",
        refractory=ref_time,
        dtype={
            "ref": defaultclock.variables["t"].dtype,
            "ref_no_unit": defaultclock.variables["t"].dtype,
        },
    )
    G.ref = 10 * ms
    G.ref_no_unit = 10
    # The neuron should spike after 5ms but then not spike for the next
    # 10ms. The state variable should continue to integrate so there should
    # be a spike after 15ms
    spike_mon = SpikeMonitor(G)
    run(16 * ms)
    assert_allclose(spike_mon.t, [5, 15] * ms)


@pytest.mark.codegen_independent
def test_refractoriness_types():
    # make sure that using a wrong type of refractoriness does not work
    group = NeuronGroup(1, "", refractory="3*Hz")
    with pytest.raises(BrianObjectException) as exc:
        Network(group).run(0 * ms)
    assert exc_isinstance(exc, TypeError)
    group = NeuronGroup(1, "ref: Hz", refractory="ref")
    with pytest.raises(BrianObjectException) as exc:
        Network(group).run(0 * ms)
    assert exc_isinstance(exc, TypeError)
    group = NeuronGroup(1, "", refractory="3")
    with pytest.raises(BrianObjectException) as exc:
        Network(group).run(0 * ms)
    assert exc_isinstance(exc, TypeError)
    group = NeuronGroup(1, "ref: 1", refractory="ref")
    with pytest.raises(BrianObjectException) as exc:
        Network(group).run(0 * ms)
    assert exc_isinstance(exc, TypeError)


@pytest.mark.codegen_independent
def test_conditional_write_set():
    """
    Test that the conditional_write attribute is set correctly
    """
    G = NeuronGroup(
        1,
        """
        dv/dt = 10*Hz : 1 (unless refractory)
        dw/dt = 10*Hz : 1
        """,
        refractory=2 * ms,
    )
    assert G.variables["v"].conditional_write is G.variables["not_refractory"]
    assert G.variables["w"].conditional_write is None


@pytest.mark.standalone_compatible
def test_conditional_write_behaviour():
    H = NeuronGroup(1, "v:1", threshold="v>-1")

    tau = 1 * ms
    eqs = """
    dv/dt = (2-v)/tau : 1 (unless refractory)
    dx/dt = 0/tau : 1 (unless refractory)
    dy/dt = 0/tau : 1
    """
    reset = """
    v = 0
    x -= 0.05
    y -= 0.05
    """
    G = NeuronGroup(1, eqs, threshold="v>1", reset=reset, refractory=1 * ms)

    Sx = Synapses(H, G, on_pre="x += dt*100*Hz")
    Sx.connect(True)

    Sy = Synapses(H, G, on_pre="y += dt*100*Hz")
    Sy.connect(True)

    M = StateMonitor(G, variables=True, record=True)

    run(10 * ms)

    assert G.x[0] < 0.2
    assert G.y[0] > 0.2
    assert G.v[0] < 1.1


@pytest.mark.standalone_compatible
def test_conditional_write_automatic_and_manual():
    source = NeuronGroup(1, "", threshold="True")  # spiking all the time
    target = NeuronGroup(
        2,
        """
        dv/dt = 0/ms : 1 (unless refractory)
        dw/dt = 0/ms : 1
        """,
        threshold="t == 0*ms",
        refractory="False",
    )  # only refractory for the first time step
    # Cell is spiking/refractory only in the first time step
    syn = Synapses(
        source,
        target,
        on_pre="""
        v += 1
        w += 1 * int(not_refractory_post)
        """,
    )
    syn.connect()
    mon = StateMonitor(target, ["v", "w"], record=True, when="end")
    run(2 * defaultclock.dt)

    # Synapse should not have been effective in the first time step
    assert_allclose(mon.v[:, 0], 0)
    assert_allclose(mon.v[:, 1], 1)
    assert_allclose(mon.w[:, 0], 0)
    assert_allclose(mon.w[:, 1], 1)


if __name__ == "__main__":
    test_add_refractoriness()
    test_missing_refractory_warning()
    test_refractoriness_basic()
    test_refractoriness_variables()
    test_refractoriness_threshold()
    test_refractoriness_threshold_basic()
    test_refractoriness_repeated()
    test_refractoriness_repeated_legacy()
    test_refractoriness_types()
    test_conditional_write_set()
    test_conditional_write_behaviour()
    test_conditional_write_automatic_and_manual()



"""
Package implementing the C++ "standalone" `Device` and `CodeObject`.
"""

from .codeobject import CPPStandaloneCodeObject
from .device import cpp_standalone_device
from .GSLcodeobject import GSLCPPStandaloneCodeObject



#!/usr/bin/env python
"""
The most simple case how to use standalone mode.
"""
from brian2 import *
set_device('cpp_standalone')  # ← only difference to "normal" simulation

tau = 10*ms
eqs = '''
dv/dt = (1-v)/tau : 1
'''
G = NeuronGroup(10, eqs, method='exact')
G.v = 'rand()'
mon = StateMonitor(G, 'v', record=True)
run(100*ms)

plt.plot(mon.t/ms, mon.v.T)
plt.gca().set(xlabel='t (ms)', ylabel='v')
plt.show()



"""
Implementation of `TimedArray`.
"""

import numpy as np

from brian2.core.clocks import defaultclock
from brian2.core.functions import Function
from brian2.core.names import Nameable
from brian2.units.allunits import second
from brian2.units.fundamentalunits import (
    Quantity,
    check_units,
    get_dimensions,
    get_unit,
)
from brian2.utils.caching import CacheKey
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import replace

__all__ = ["TimedArray"]


logger = get_logger(__name__)


def _find_K(group_dt, dt):
    dt_ratio = dt / group_dt
    if dt_ratio > 1 and np.floor(dt_ratio) != dt_ratio:
        logger.warn(
            "Group uses a dt of %s while TimedArray uses dt "
            "of %s (ratio: 1/%s) → time grids not aligned"
            % (group_dt * second, dt * second, dt_ratio),
            once=True,
        )
    # Find an upsampling factor that should avoid rounding issues even
    # for multistep methods
    K = max(int(2 ** np.ceil(np.log2(8 / group_dt * dt))), 1)
    return K


def _generate_cpp_code_1d(values, dt, name):
    def cpp_impl(owner):
        K = _find_K(owner.clock.dt_, dt)
        code = (
            """
        static inline double %NAME%(const double t)
        {
            const double epsilon = %DT% / %K%;
            int i = (int)((t/epsilon + 0.5)/%K%);
            if(i < 0)
               i = 0;
            if(i >= %NUM_VALUES%)
                i = %NUM_VALUES%-1;
            return _namespace%NAME%_values[i];
        }
        """.replace(
                "%NAME%", name
            )
            .replace("%DT%", f"{dt:.18f}")
            .replace("%K%", str(K))
            .replace("%NUM_VALUES%", str(len(values)))
        )

        return code

    return cpp_impl


def _generate_cpp_code_2d(values, dt, name):
    def cpp_impl(owner):
        K = _find_K(owner.clock.dt_, dt)
        support_code = """
        static inline double %NAME%(const double t, const int i)
        {
            const double epsilon = %DT% / %K%;
            if (i < 0 || i >= %COLS%)
                return NAN;
            int timestep = (int)((t/epsilon + 0.5)/%K%);
            if(timestep < 0)
               timestep = 0;
            else if(timestep >= %ROWS%)
                timestep = %ROWS%-1;
            return _namespace%NAME%_values[timestep*%COLS% + i];
        }
        """
        code = replace(
            support_code,
            {
                "%NAME%": name,
                "%DT%": f"{dt:.18f}",
                "%K%": str(K),
                "%COLS%": str(values.shape[1]),
                "%ROWS%": str(values.shape[0]),
            },
        )
        return code

    return cpp_impl


def _generate_cython_code_1d(values, dt, name):
    def cython_impl(owner):
        K = _find_K(owner.clock.dt_, dt)
        code = (
            """
        cdef double %NAME%(const double t):
            global _namespace%NAME%_values
            cdef double epsilon = %DT% / %K%
            cdef int i = (int)((t/epsilon + 0.5)/%K%)
            if i < 0:
               i = 0
            if i >= %NUM_VALUES%:
                i = %NUM_VALUES% - 1
            return _namespace%NAME%_values[i]
        """.replace(
                "%NAME%", name
            )
            .replace("%DT%", f"{dt:.18f}")
            .replace("%K%", str(K))
            .replace("%NUM_VALUES%", str(len(values)))
        )

        return code

    return cython_impl


def _generate_cython_code_2d(values, dt, name):
    def cython_impl(owner):
        K = _find_K(owner.clock.dt_, dt)
        code = """
        cdef double %NAME%(const double t, const int i):
            global _namespace%NAME%_values
            cdef double epsilon = %DT% / %K%
            if i < 0 or i >= %COLS%:
                return _numpy.nan
            cdef int timestep = (int)((t/epsilon + 0.5)/%K%)
            if timestep < 0:
               timestep = 0
            elif timestep >= %ROWS%:
                timestep = %ROWS%-1
            return _namespace%NAME%_values[timestep*%COLS% + i]
        """
        code = replace(
            code,
            {
                "%NAME%": name,
                "%DT%": f"{dt:.18f}",
                "%K%": str(K),
                "%COLS%": str(values.shape[1]),
                "%ROWS%": str(values.shape[0]),
            },
        )
        return code

    return cython_impl


class TimedArray(Function, Nameable, CacheKey):
    """
    TimedArray(values, dt, name=None)

    A function of time built from an array of values. The returned object can
    be used as a function, including in model equations etc. The resulting
    function has to be called as `funcion_name(t)` if the provided value array
    is one-dimensional and as `function_name(t, i)` if it is two-dimensional.

    Parameters
    ----------
    values : ndarray or `Quantity`
        An array of values providing the values at various points in time. This
        array can either be one- or two-dimensional. If it is two-dimensional
        it's first dimension should be the time.
    dt : `Quantity`
        The time distance between values in the `values` array.
    name : str, optional
        A unique name for this object, see `Nameable` for details. Defaults
        to ``'_timedarray*'``.

    Notes
    -----
    For time values corresponding to elements outside of the range of `values`
    provided, the first respectively last element is returned.

    Examples
    --------
    >>> from brian2 import *
    >>> ta = TimedArray([1, 2, 3, 4] * mV, dt=0.1*ms)
    >>> print(ta(0.3*ms))
    4. mV
    >>> G = NeuronGroup(1, 'v = ta(t) : volt')
    >>> mon = StateMonitor(G, 'v', record=True)
    >>> net = Network(G, mon)
    >>> net.run(1*ms)  # doctest: +ELLIPSIS
    ...
    >>> print(mon[0].v)
    [ 1.  2.  3.  4.  4.  4.  4.  4.  4.  4.] mV
    >>> ta2d = TimedArray([[1, 2], [3, 4], [5, 6]]*mV, dt=0.1*ms)
    >>> G = NeuronGroup(4, 'v = ta2d(t, i%2) : volt')
    >>> mon = StateMonitor(G, 'v', record=True)
    >>> net = Network(G, mon)
    >>> net.run(0.2*ms)  # doctest: +ELLIPSIS
    ...
    >>> print(mon.v[:])
    [[ 1.  3.]
     [ 2.  4.]
     [ 1.  3.]
     [ 2.  4.]] mV
    """

    _cache_irrelevant_attributes = {"_id", "values", "pyfunc", "implementations"}

    #: Container for implementing functions for different targets
    #: This container can be extended by other codegeneration targets/devices
    #: The key has to be the name of the target, the value is a tuple of
    #: functions, the first for a 1d array, the second for a 2d array.
    #: The functions have to take three parameters: (values, dt, name), i.e. the
    #: array values, their physical dimensions, the dt of the TimedArray, and
    #: the name of the TimedArray. The functions have to return *a function*
    #: that takes the `owner` argument (out of which they can get the context's
    #: dt as `owner.clock.dt_`) and returns the code.
    implementations = {
        "cpp": (_generate_cpp_code_1d, _generate_cpp_code_2d),
        "cython": (_generate_cython_code_1d, _generate_cython_code_2d),
    }

    @check_units(dt=second)
    def __init__(self, values, dt, name=None):
        if name is None:
            name = "_timedarray*"
        Nameable.__init__(self, name)
        dimensions = get_dimensions(values)
        self.dim = dimensions
        values = np.asarray(values, dtype=np.float64)
        self.values = values
        dt = float(dt)
        self.dt = dt
        if values.ndim == 1:
            self._init_1d()
        elif values.ndim == 2:
            self._init_2d()
        else:
            raise NotImplementedError(
                "Only 1d and 2d arrays are supported for TimedArray"
            )

    def _init_1d(self):
        dimensions = self.dim
        unit = get_unit(dimensions)
        values = self.values
        dt = self.dt

        # Python implementation (with units), used when calling the TimedArray
        # directly, outside of a simulation
        @check_units(t=second, result=unit)
        def timed_array_func(t):
            # We round according to the current defaultclock.dt
            K = _find_K(float(defaultclock.dt), dt)
            epsilon = dt / K
            i = np.clip(
                np.int_(np.round(np.asarray(t / epsilon)) / K), 0, len(values) - 1
            )
            return Quantity(values[i], dim=dimensions)

        Function.__init__(self, pyfunc=timed_array_func)

        # we use dynamic implementations because we want to do upsampling
        # in a way that avoids rounding problems with the group's dt
        def create_numpy_implementation(owner):
            group_dt = owner.clock.dt_

            K = _find_K(group_dt, dt)
            n_values = len(values)
            epsilon = dt / K

            def unitless_timed_array_func(t):
                timestep = np.clip(np.int_(np.round(t / epsilon) / K), 0, n_values - 1)
                return values[timestep]

            unitless_timed_array_func._arg_units = [second]
            unitless_timed_array_func._return_unit = unit

            return unitless_timed_array_func

        self.implementations.add_dynamic_implementation(
            "numpy", create_numpy_implementation
        )
        namespace = lambda owner: {f"{self.name}_values": self.values}

        for target, (func_1d, _) in TimedArray.implementations.items():
            self.implementations.add_dynamic_implementation(
                target,
                func_1d(self.values, self.dt, self.name),
                namespace=namespace,
                name=self.name,
            )

    def _init_2d(self):
        dimensions = self.dim
        unit = get_unit(dimensions)
        values = self.values
        dt = self.dt

        # Python implementation (with units), used when calling the TimedArray
        # directly, outside of a simulation
        @check_units(i=1, t=second, result=unit)
        def timed_array_func(t, i):
            # We round according to the current defaultclock.dt
            K = _find_K(float(defaultclock.dt), dt)
            epsilon = dt / K
            time_step = np.clip(
                np.int_(np.round(np.asarray(t / epsilon)) / K), 0, len(values) - 1
            )
            return Quantity(values[time_step, i], dim=dimensions)

        Function.__init__(self, pyfunc=timed_array_func)

        # we use dynamic implementations because we want to do upsampling
        # in a way that avoids rounding problems with the group's dt
        def create_numpy_implementation(owner):
            group_dt = owner.clock.dt_

            K = _find_K(group_dt, dt)
            n_values = len(values)
            epsilon = dt / K

            def unitless_timed_array_func(t, i):
                timestep = np.clip(np.int_(np.round(t / epsilon) / K), 0, n_values - 1)
                return values[timestep, i]

            unitless_timed_array_func._arg_units = [second]
            unitless_timed_array_func._return_unit = unit

            return unitless_timed_array_func

        self.implementations.add_dynamic_implementation(
            "numpy", create_numpy_implementation
        )
        values_flat = self.values.astype(np.double, order="C", copy=False).ravel()
        namespace = lambda owner: {f"{self.name}_values": values_flat}

        for target, (_, func_2d) in TimedArray.implementations.items():
            self.implementations.add_dynamic_implementation(
                target,
                func_2d(self.values, self.dt, self.name),
                namespace=namespace,
                name=self.name,
            )

    def is_locally_constant(self, dt):
        if dt > self.dt:
            return False
        dt_ratio = self.dt / float(dt)
        if np.floor(dt_ratio) != dt_ratio:
            logger.info(
                "dt of the TimedArray is not an integer multiple of "
                "the group's dt, the TimedArray's return value can "
                "therefore not be considered constant over one "
                "timestep, making exact integration impossible.",
                once=True,
            )
            return False
        return True



import os

import mayavi.mlab as mayavi
from brian2 import *

def find_max_coordinates(morpho, current_max):
    max_x, max_y, max_z = np.max(np.abs(morpho.coordinates), axis=0)
    new_max = (max([max_x, current_max[0]]),
               max([max_y, current_max[1]]),
               max([max_z, current_max[2]]))
    for child in morpho.children:
        new_max = find_max_coordinates(child, new_max)
    return new_max

DARKRED = (0.5450980392156862, 0.0, 0.0)
DARKBLUE = (0.0, 0.0, 0.5450980392156862)

def plot_morphology3D(morpho, color_switch=False, show_compartments=False):
    if isinstance(morpho, Soma):
        mayavi.points3d(morpho.x/um, morpho.y/um, morpho.z/um, morpho.diameter/um,
                        color=DARKRED, scale_factor=1.0, resolution=16)
    else:
        coords = morpho.coordinates
        if color_switch:
            color = DARKBLUE
        else:
            color = DARKRED
        mayavi.plot3d(coords[:, 0]/um, coords[:, 1]/um, coords[:, 2]/um, color=color,
                      tube_radius=1)
        # dots at the center of the compartments
        if show_compartments:
            mayavi.points3d(coords[:, 0]/um, coords[:, 1]/um, coords[:, 2]/um,
                            np.ones(coords.shape[0]), color=color, scale_factor=1,
                            transparent=0.25)

    for child in morpho.children:
        plot_morphology3D(child, color_switch=not color_switch)

def plot_morphology2D(morpho, axis, color_switch=False):
    if isinstance(morpho, Soma):
        ax.plot(morpho.x/um, morpho.y/um, 'o', color='darkred',
                ms=morpho.diameter/um, mec='none')
    else:
        coords = morpho.coordinates
        if color_switch:
            color = 'darkblue'
        else:
            color = 'darkred'
        ax.plot(coords[:, 0]/um, coords[:, 1]/um, color='black',
                lw=2)
        # dots at the center of the compartments
        ax.plot(morpho.x/um, morpho.y/um, 'o', color=color,
                mec='none', alpha=0.75)

    for child in morpho.children:
        plot_morphology2D(child, axis=axis, color_switch=not color_switch)

if __name__ == '__main__':
    dirname = os.path.dirname(__file__)
    PATH = os.path.join(dirname, '..', '..', '..', 'docs_sphinx', 'user', 'images')

    # Construct binary tree according to Rall's formula
    diameter = 32*um
    length = 80*um
    morpho = Soma(diameter=diameter)
    endpoints = {morpho}
    for depth in range(1, 10):
        diameter /= 2.**(1./3.)
        length /= 2.**(2./3.)
        new_endpoints = set()
        for endpoint in endpoints:
            new_L = Cylinder(n=max([1, int(length/(5*um))]), diameter=diameter,
                             length=length)
            new_R = Cylinder(n=max([1, int(length/(5*um))]), diameter=diameter,
                             length=length)
            new_endpoints.add(new_L)
            new_endpoints.add(new_R)
            endpoint.L = new_L
            endpoint.R = new_R
        endpoints = new_endpoints
    print('total number of sections and compartments', morpho.n_sections, len(morpho))
    morpho_with_coords = morpho.generate_coordinates()

    ax = plt.subplot(111)
    plot_morphology2D(morpho_with_coords, ax)
    plt.axis('equal')
    plt.xlabel('x ($\mu$ m)')
    plt.ylabel('y ($\mu$ m)')

    plt.savefig(os.path.join(PATH, 'morphology_deterministic_coords.png'))

    print('be careful, this plotting takes a long time')
    for title, noise_sec, noise_comp in [('section', 25, 0),
                                         ('section_compartment', 25, 15)]:

        for idx in range(3):
            fig = mayavi.figure(bgcolor=(0.95, 0.95, 0.95))
            print(idx)
            morpho_with_coords = morpho.generate_coordinates(section_randomness=noise_sec,
                                                 compartment_randomness=noise_comp)
            plot_morphology3D(morpho_with_coords)
            cam = fig.scene.camera
            cam.zoom(1.1)
            mayavi.draw()
            mayavi.savefig(os.path.join(PATH, 'morphology_random_%s_%d.png' % (title, idx+1)))
            mayavi.close()



"""
Differential equations for Brian models.
"""

import keyword
import re
import string
from collections import namedtuple
from collections.abc import Hashable, Mapping

import sympy
from pyparsing import (
    CharsNotIn,
    Combine,
    Group,
    LineEnd,
    OneOrMore,
    Optional,
    ParseException,
    Suppress,
    Word,
    ZeroOrMore,
    restOfLine,
)

from brian2.core.namespace import DEFAULT_CONSTANTS, DEFAULT_FUNCTIONS, DEFAULT_UNITS
from brian2.parsing.sympytools import str_to_sympy, sympy_to_str
from brian2.units.allunits import second
from brian2.units.fundamentalunits import (
    DIMENSIONLESS,
    DimensionMismatchError,
    Quantity,
    Unit,
    get_dimensions,
    get_unit,
    get_unit_for_display,
)
from brian2.utils.caching import CacheKey, cached
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import get_identifiers
from brian2.utils.topsort import topsort

from .codestrings import Expression
from .unitcheck import check_dimensions

__all__ = ["Equations"]

logger = get_logger(__name__)

# Equation types (currently simple strings but always use the constants,
# this might get refactored into objects, for example)
PARAMETER = "parameter"
DIFFERENTIAL_EQUATION = "differential equation"
SUBEXPRESSION = "subexpression"

# variable types (FLOAT is the only one that is possible for variables that
# have dimensions). These types will be later translated into dtypes, either
# using the default values from the preferences, or explicitly given dtypes in
# the construction of the `NeuronGroup`, `Synapses`, etc. object
FLOAT = "float"
INTEGER = "integer"
BOOLEAN = "boolean"

# Definitions of equation structure for parsing with pyparsing
# TODO: Maybe move them somewhere else to not pollute the namespace here?
#       Only IDENTIFIER and EQUATIONS are ever used later
###############################################################################
# Basic Elements
###############################################################################

# identifiers like in C: can start with letter or underscore, then a
# combination of letters, numbers and underscores
# Note that the check_identifiers function later performs more checks, e.g.
# names starting with underscore should only be used internally
IDENTIFIER = Word(
    string.ascii_letters + "_", string.ascii_letters + string.digits + "_"
).setResultsName("identifier")

# very broad definition here, expression will be analysed by sympy anyway
# allows for multi-line expressions, where each line can have comments
EXPRESSION = Combine(
    OneOrMore(
        (CharsNotIn(":#\n") + Suppress(Optional(LineEnd()))).ignore("#" + restOfLine)
    ),
    joinString=" ",
).setResultsName("expression")

# a unit
# very broad definition here, again. Whether this corresponds to a valid unit
# string will be checked later
UNIT = Word(string.ascii_letters + string.digits + "*/.- ").setResultsName("unit")

# a single Flag (e.g. "const" or "event-driven")
FLAG = Word(string.ascii_letters, string.ascii_letters + "_- " + string.digits)

# Flags are comma-separated and enclosed in parantheses: "(flag1, flag2)"
FLAGS = (
    Suppress("(") + FLAG + ZeroOrMore(Suppress(",") + FLAG) + Suppress(")")
).setResultsName("flags")

###############################################################################
# Equations
###############################################################################
# Three types of equations
# Parameter:
# x : volt (flags)
PARAMETER_EQ = Group(
    IDENTIFIER + Suppress(":") + UNIT + Optional(FLAGS)
).setResultsName(PARAMETER)

# Static equation:
# x = 2 * y : volt (flags)
STATIC_EQ = Group(
    IDENTIFIER + Suppress("=") + EXPRESSION + Suppress(":") + UNIT + Optional(FLAGS)
).setResultsName(SUBEXPRESSION)

# Differential equation
# dx/dt = -x / tau : volt
DIFF_OP = Suppress("d") + IDENTIFIER + Suppress("/") + Suppress("dt")
DIFF_EQ = Group(
    DIFF_OP + Suppress("=") + EXPRESSION + Suppress(":") + UNIT + Optional(FLAGS)
).setResultsName(DIFFERENTIAL_EQUATION)

# ignore comments
EQUATION = (PARAMETER_EQ | STATIC_EQ | DIFF_EQ).ignore("#" + restOfLine)
EQUATIONS = ZeroOrMore(EQUATION)


class EquationError(Exception):
    """
    Exception type related to errors in an equation definition.
    """

    pass


def check_identifier_basic(identifier):
    """
    Check an identifier (usually resulting from an equation string provided by
    the user) for conformity with the rules. The rules are:

        1. Only ASCII characters
        2. Starts with a character, then mix of alphanumerical characters and
           underscore
        3. Is not a reserved keyword of Python

    Parameters
    ----------
    identifier : str
        The identifier that should be checked

    Raises
    ------
    SyntaxError
        If the identifier does not conform to the above rules.
    """

    # Check whether the identifier is parsed correctly -- this is always the
    # case, if the identifier results from the parsing of an equation but there
    # might be situations where the identifier is specified directly
    parse_result = list(IDENTIFIER.scanString(identifier))

    # parse_result[0][0][0] refers to the matched string -- this should be the
    # full identifier, if not it is an illegal identifier like "3foo" which only
    # matched on "foo"
    if len(parse_result) != 1 or parse_result[0][0][0] != identifier:
        raise SyntaxError(f"'{identifier}' is not a valid variable name.")

    if keyword.iskeyword(identifier):
        raise SyntaxError(
            f"'{identifier}' is a Python keyword and cannot be used as a variable."
        )

    if identifier.startswith("_"):
        raise SyntaxError(
            f"Variable '{identifier}' starts with an underscore, "
            "this is only allowed for variables used "
            "internally"
        )


def check_identifier_reserved(identifier):
    """
    Check that an identifier is not using a reserved special variable name. The
    special variables are: 't', 'dt', and 'xi', as well as everything starting
    with `xi_`.

    Parameters
    ----------
    identifier: str
        The identifier that should be checked

    Raises
    ------
    SyntaxError
        If the identifier is a special variable name.
    """
    if identifier in (
        "t",
        "dt",
        "t_in_timesteps",
        "xi",
        "i",
        "N",
    ) or identifier.startswith("xi_"):
        raise SyntaxError(
            f"'{identifier}' has a special meaning in equations and "
            "cannot be used as a variable name."
        )


def check_identifier_units(identifier):
    """
    Make sure that identifier names do not clash with unit names.
    """
    if identifier in DEFAULT_UNITS:
        raise SyntaxError(
            f"'{identifier}' is the name of a unit, cannot be used as a variable name."
        )


def check_identifier_functions(identifier):
    """
    Make sure that identifier names do not clash with function names.
    """
    if identifier in DEFAULT_FUNCTIONS:
        raise SyntaxError(
            f"'{identifier}' is the name of a function, cannot be used as "
            "a variable name."
        )


def check_identifier_constants(identifier):
    """
    Make sure that identifier names do not clash with function names.
    """
    if identifier in DEFAULT_CONSTANTS:
        raise SyntaxError(
            f"'{identifier}' is the name of a constant, cannot be used as "
            "a variable name."
        )


_base_units_with_alternatives = None
_base_units = None


def dimensions_and_type_from_string(unit_string):
    """
    Returns the physical dimensions that results from evaluating a string like
    "siemens / metre ** 2", allowing for the special string "1" to signify
    dimensionless units, the string "boolean" for a boolean and "integer" for
    an integer variable.

    Parameters
    ----------
    unit_string : str
        The string that should evaluate to a unit

    Returns
    -------
    d, type : (`Dimension`, {FLOAT, INTEGER or BOOL})
        The resulting physical dimensions and the type of the variable.

    Raises
    ------
    ValueError
        If the string cannot be evaluated to a unit.
    """
    # Lazy import to avoid circular dependency
    from brian2.core.namespace import DEFAULT_UNITS

    global _base_units_with_alternatives
    global _base_units
    if _base_units_with_alternatives is None:
        base_units_for_dims = {}
        for unit_name, unit in reversed(DEFAULT_UNITS.items()):
            if float(unit) == 1.0 and repr(unit)[-1] not in ["2", "3"]:
                if unit.dim in base_units_for_dims:
                    if unit_name not in base_units_for_dims[unit.dim]:
                        base_units_for_dims[unit.dim].append(unit_name)
                else:
                    base_units_for_dims[unit.dim] = [repr(unit)]
                    if unit_name != repr(unit):
                        base_units_for_dims[unit.dim].append(unit_name)
        alternatives = sorted(
            [tuple(values) for values in base_units_for_dims.values()]
        )
        _base_units = {v: DEFAULT_UNITS[v] for values in alternatives for v in values}
        # Create a string that lists all allowed base units
        alternative_strings = []
        for units in alternatives:
            string = units[0]
            if len(units) > 1:
                other_units = ", ".join(units[1:])
                string += f" ({other_units})"
            alternative_strings.append(string)
        _base_units_with_alternatives = ", ".join(alternative_strings)

    unit_string = unit_string.strip()

    # Special case: dimensionless unit
    if unit_string == "1":
        return DIMENSIONLESS, FLOAT

    # Another special case: boolean variable
    if unit_string == "boolean":
        return DIMENSIONLESS, BOOLEAN
    if unit_string == "bool":
        raise TypeError("Use 'boolean' not 'bool' as the unit for a boolean variable.")

    # Yet another special case: integer variable
    if unit_string == "integer":
        return DIMENSIONLESS, INTEGER

    # Check first whether the expression only refers to base units
    identifiers = get_identifiers(unit_string)
    for identifier in identifiers:
        if identifier not in _base_units:
            if identifier in DEFAULT_UNITS:
                # A known unit, but not a base unit
                base_unit = get_unit(DEFAULT_UNITS[identifier].dim)
                if not repr(base_unit) in _base_units:
                    # Make sure that we don't suggest a unit that is not allowed
                    # (should not happen, normally)
                    base_unit = Unit(1, dim=base_unit.dim)
                raise ValueError(
                    "Unit specification refers to "
                    f"'{identifier}', but this is not a base "
                    f"unit. Use '{base_unit!r}' instead."
                )
            else:
                # Not a known unit
                raise ValueError(
                    "Unit specification refers to "
                    f"'{identifier}', but this is not a base "
                    "unit. The following base units are "
                    f"allowed: {_base_units_with_alternatives}."
                )
    try:
        evaluated_unit = eval(unit_string, _base_units)
    except Exception as ex:
        raise ValueError(
            f"Could not interpret '{unit_string}' as a unit specification: {ex}"
        )

    # Check whether the result is a unit
    if not isinstance(evaluated_unit, Unit):
        if isinstance(evaluated_unit, Quantity):
            raise ValueError(
                f"'{unit_string}' does not evaluate to a unit but to a "
                "quantity -- make sure to only use units, e.g. "
                "'siemens/metre**2' and not '1 * siemens/metre**2'"
            )
        else:
            raise ValueError(
                f"'{unit_string}' does not evaluate to a unit, the result "
                f"has type {type(evaluated_unit)} instead."
            )

    # No error has been raised, all good
    return evaluated_unit.dim, FLOAT


@cached
def parse_string_equations(eqns):
    """
    parse_string_equations(eqns)

    Parse a string defining equations.

    Parameters
    ----------
    eqns : str
        The (possibly multi-line) string defining the equations. See the
        documentation of the `Equations` class for details.

    Returns
    -------
    equations : dict
        A dictionary mapping variable names to
        `~brian2.equations.equations.Equations` objects
    """
    equations = {}

    try:
        parsed = EQUATIONS.parseString(eqns, parseAll=True)
    except ParseException as p_exc:
        raise EquationError(
            "Parsing failed: \n"
            + str(p_exc.line)
            + "\n"
            + " " * (p_exc.column - 1)
            + "^\n"
            + str(p_exc)
        ) from p_exc
    for eq in parsed:
        eq_type = eq.getName()
        eq_content = dict(eq.items())
        # Check for reserved keywords
        identifier = eq_content["identifier"]

        # Convert unit string to Unit object
        try:
            dims, var_type = dimensions_and_type_from_string(eq_content["unit"])
        except ValueError as ex:
            raise EquationError(
                "Error parsing the unit specification for "
                f"variable '{identifier}': {ex}"
            )

        expression = eq_content.get("expression", None)
        if expression is not None:
            # Replace multiple whitespaces (arising from joining multiline
            # strings) with single space
            p = re.compile(r"\s{2,}")
            expression = Expression(p.sub(" ", expression))
        flags = list(eq_content.get("flags", []))

        equation = SingleEquation(
            eq_type, identifier, dims, var_type=var_type, expr=expression, flags=flags
        )

        if identifier in equations:
            raise EquationError(f"Duplicate definition of variable '{identifier}'")

        equations[identifier] = equation

    return equations


class SingleEquation(Hashable, CacheKey):
    """
    Class for internal use, encapsulates a single equation or parameter.

    .. note::
        This class should never be used directly, it is only useful as part of
        the `Equations` class.

    Parameters
    ----------
    type : {PARAMETER, DIFFERENTIAL_EQUATION, SUBEXPRESSION}
        The type of the equation.
    varname : str
        The variable that is defined by this equation.
    dimensions : `Dimension`
        The physical dimensions of the variable
    var_type : {FLOAT, INTEGER, BOOLEAN}
        The type of the variable (floating point value or boolean).
    expr : `Expression`, optional
        The expression defining the variable (or ``None`` for parameters).
    flags: list of str, optional
        A list of flags that give additional information about this equation.
        What flags are possible depends on the type of the equation and the
        context.
    """

    _cache_irrelevant_attributes = {"update_order"}

    def __init__(
        self, type, varname, dimensions, var_type=FLOAT, expr=None, flags=None
    ):
        self.type = type
        self.varname = varname
        self.dim = get_dimensions(dimensions)
        self.var_type = var_type
        if dimensions is not DIMENSIONLESS:
            if var_type == BOOLEAN:
                raise TypeError("Boolean variables are necessarily dimensionless.")
            elif var_type == INTEGER:
                raise TypeError("Integer variables are necessarily dimensionless.")

        if type == DIFFERENTIAL_EQUATION:
            if var_type != FLOAT:
                raise TypeError(
                    "Differential equations can only define floating point variables"
                )
        self.expr = expr
        if flags is None:
            self.flags = []
        else:
            self.flags = list(flags)

        # will be set later in the sort_subexpressions method of Equations
        self.update_order = -1

    unit = property(lambda self: get_unit(self.dim), doc="The `Unit` of this equation.")

    identifiers = property(
        lambda self: self.expr.identifiers if self.expr is not None else set(),
        doc="All identifiers in the RHS of this equation.",
    )

    stochastic_variables = property(
        lambda self: {
            variable
            for variable in self.identifiers
            if variable == "xi" or variable.startswith("xi_")
        },
        doc="Stochastic variables in the RHS of this equation",
    )

    def __eq__(self, other):
        if not isinstance(other, SingleEquation):
            return NotImplemented
        return self._state_tuple == other._state_tuple

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._state_tuple)

    def _latex(self, *args):
        if self.type == DIFFERENTIAL_EQUATION:
            return (
                r"\frac{\mathrm{d}"
                + sympy.latex(sympy.Symbol(self.varname))
                + r"}{\mathrm{d}t} = "
                + sympy.latex(str_to_sympy(self.expr.code))
            )
        elif self.type == SUBEXPRESSION:
            return (
                sympy.latex(sympy.Symbol(self.varname))
                + " = "
                + sympy.latex(str_to_sympy(self.expr.code))
            )
        elif self.type == PARAMETER:
            return sympy.latex(sympy.Symbol(self.varname))

    def __str__(self):
        if self.type == DIFFERENTIAL_EQUATION:
            s = "d" + self.varname + "/dt"
        else:
            s = self.varname

        if self.expr is not None:
            s += " = " + str(self.expr)

        s += " : " + get_unit_for_display(self.dim)

        if len(self.flags):
            s += " (" + ", ".join(self.flags) + ")"

        return s

    def __repr__(self):
        s = "<" + self.type + " " + self.varname

        if self.expr is not None:
            s += ": " + self.expr.code

        s += " (Unit: " + get_unit_for_display(self.dim)

        if len(self.flags):
            s += ", flags: " + ", ".join(self.flags)

        s += ")>"
        return s

    def _repr_pretty_(self, p, cycle):
        """
        Pretty printing for ipython.
        """
        if cycle:
            # should never happen
            raise AssertionError("Cyclical call of SingleEquation._repr_pretty")

        if self.type == DIFFERENTIAL_EQUATION:
            p.text("d" + self.varname + "/dt")
        else:
            p.text(self.varname)

        if self.expr is not None:
            p.text(" = ")
            p.pretty(self.expr)

        p.text(" : ")
        p.pretty(get_unit(self.dim))

        if len(self.flags):
            p.text(" (" + ", ".join(self.flags) + ")")

    def _repr_latex_(self):
        return "$" + sympy.latex(self) + "$"


class Equations(Hashable, Mapping):
    """
    Container that stores equations from which models can be created.

    String equations can be of any of the following forms:

    1. ``dx/dt = f : unit (flags)`` (differential equation)
    2. ``x = f : unit (flags)`` (equation)
    3. ``x : unit (flags)`` (parameter)

    String equations can span several lines and contain Python-style comments
    starting with ``#``

    Parameters
    ----------
    eqs : `str` or list of `SingleEquation` objects
        A multiline string of equations (see above) -- for internal purposes
        also a list of `SingleEquation` objects can be given. This is done for
        example when adding new equations to implement the refractory
        mechanism. Note that in this case the variable names are not checked
        to allow for "internal names", starting with an underscore.
    kwds: keyword arguments
        Keyword arguments can be used to replace variables in the equation
        string. Arguments have to be of the form ``varname=replacement``, where
        `varname` has to correspond to a variable name in the given equation.
        The replacement can be either a string (replacing a name with a new
        name, e.g. ``tau='tau_e'``) or a value (replacing the variable name
        with the value, e.g. ``tau=tau_e`` or ``tau=10*ms``).
    """

    def __init__(self, eqns, **kwds):
        if isinstance(eqns, str):
            self._equations = parse_string_equations(eqns)
            # Do a basic check for the identifiers
            self.check_identifiers()
        else:
            self._equations = {}
            for eq in eqns:
                if not isinstance(eq, SingleEquation):
                    raise TypeError(
                        "The list should only contain "
                        f"SingleEquation objects, not {type(eq)}"
                    )
                if eq.varname in self._equations:
                    raise EquationError(
                        f"Duplicate definition of variable '{eq.varname}'"
                    )
                self._equations[eq.varname] = eq

        self._equations = self._substitute(kwds)

        # Check for special symbol xi (stochastic term)
        uses_xi = None
        for eq in self._equations.values():
            if eq.expr is not None and "xi" in eq.expr.identifiers:
                if not eq.type == DIFFERENTIAL_EQUATION:
                    raise EquationError(
                        f"The equation defining '{eq.varname}' "
                        "contains the symbol 'xi' but is not a "
                        "differential equation."
                    )
                elif uses_xi is not None:
                    raise EquationError(
                        f"The equation defining {eq.varname} contains "
                        "the symbol 'xi', but it is already used "
                        f"in the equation defining {uses_xi}. Rename "
                        "the variables to 'xi_...' to make "
                        "clear whether they are the same or "
                        "independent random variables. Using "
                        "the same name twice will lead to "
                        "identical noise realizations "
                        "whereas using different names will "
                        "lead to independent noise "
                        "realizations."
                    )
                else:
                    uses_xi = eq.varname

        # rearrange subexpressions
        self._sort_subexpressions()

        #: Cache for equations with the subexpressions substituted
        self._substituted_expressions = None

    def _substitute(self, replacements):
        if len(replacements) == 0:
            return self._equations

        new_equations = {}
        for eq in self.values():
            # Replace the name of a model variable (works only for strings)
            if eq.varname in replacements:
                new_varname = replacements[eq.varname]
                if not isinstance(new_varname, str):
                    raise ValueError(
                        f"Cannot replace model variable '{eq.varname}' with a value."
                    )
                if new_varname in self or new_varname in new_equations:
                    raise EquationError(
                        f"Cannot replace model variable '{eq.varname}' "
                        f"with '{new_varname}', duplicate definition "
                        f"of '{new_varname}'."
                    )
                # make sure that the replacement is a valid identifier
                Equations.check_identifier(new_varname)
            else:
                new_varname = eq.varname

            if eq.type in [SUBEXPRESSION, DIFFERENTIAL_EQUATION]:
                # Replace values in the RHS of the equation
                new_code = eq.expr.code
                for to_replace, replacement in replacements.items():
                    if to_replace in eq.identifiers:
                        if isinstance(replacement, str):
                            # replace the name with another name
                            new_code = re.sub(
                                "\\b" + to_replace + "\\b", replacement, new_code
                            )
                        else:
                            # replace the name with a value
                            new_code = re.sub(
                                "\\b" + to_replace + "\\b",
                                "(" + repr(replacement) + ")",
                                new_code,
                            )
                        try:
                            Expression(new_code)
                        except ValueError as ex:
                            raise ValueError(
                                'Replacing "%s" with "%r" failed: %s'
                                % (to_replace, replacement, ex)
                            )
                new_equations[new_varname] = SingleEquation(
                    eq.type,
                    new_varname,
                    dimensions=eq.dim,
                    var_type=eq.var_type,
                    expr=Expression(new_code),
                    flags=eq.flags,
                )
            else:
                new_equations[new_varname] = SingleEquation(
                    eq.type,
                    new_varname,
                    dimensions=eq.dim,
                    var_type=eq.var_type,
                    flags=eq.flags,
                )

        return new_equations

    def substitute(self, **kwds):
        return Equations(list(self._substitute(kwds).values()))

    def __iter__(self):
        return iter(self._equations)

    def __len__(self):
        return len(self._equations)

    def __getitem__(self, key):
        return self._equations[key]

    def __add__(self, other_eqns):
        if isinstance(other_eqns, str):
            other_eqns = parse_string_equations(other_eqns)
        elif not isinstance(other_eqns, Equations):
            return NotImplemented

        return Equations(list(self.values()) + list(other_eqns.values()))

    def __hash__(self):
        return hash(frozenset(self._equations.items()))

    #: A set of functions that are used to check identifiers (class attribute).
    #: Functions can be registered with the static method
    #: `Equations.register_identifier_check` and will be automatically
    #: used when checking identifiers
    identifier_checks = {
        check_identifier_basic,
        check_identifier_reserved,
        check_identifier_functions,
        check_identifier_constants,
        check_identifier_units,
    }

    @staticmethod
    def register_identifier_check(func):
        """
        Register a function for checking identifiers.

        Parameters
        ----------
        func : callable
            The function has to receive a single argument, the name of the
            identifier to check, and raise a ValueError if the identifier
            violates any rule.

        """
        if not callable(func):
            raise ValueError("Can only register callables.")

        Equations.identifier_checks.add(func)

    @staticmethod
    def check_identifier(identifier):
        """
        Perform all the registered checks. Checks can be registered via
        `Equations.register_identifier_check`.

        Parameters
        ----------
        identifier : str
            The identifier that should be checked

        Raises
        ------
        ValueError
            If any of the registered checks fails.
        """
        for check_func in Equations.identifier_checks:
            check_func(identifier)

    def check_identifiers(self):
        """
        Check all identifiers for conformity with the rules.

        Raises
        ------
        ValueError
            If an identifier does not conform to the rules.

        See also
        --------
        Equations.check_identifier : The function that is called for each identifier.
        """
        for name in self.names:
            Equations.check_identifier(name)

    def get_substituted_expressions(self, variables=None, include_subexpressions=False):
        """
        Return a list of ``(varname, expr)`` tuples, containing all
        differential equations (and optionally subexpressions) with all the
        subexpression variables substituted with the respective expressions.

        Parameters
        ----------
        variables : dict, optional
            A mapping of variable names to `Variable`/`Function` objects.
        include_subexpressions : bool
            Whether also to return substituted subexpressions. Defaults to
            ``False``.

        Returns
        -------
        expr_tuples : list of (str, `CodeString`)
            A list of ``(varname, expr)`` tuples, where ``expr`` is a
            `CodeString` object with all subexpression variables substituted
            with the respective expression.
        """
        if self._substituted_expressions is None:
            self._substituted_expressions = []
            substitutions = {}
            for eq in self.ordered:
                # Skip parameters
                if eq.expr is None:
                    continue

                new_sympy_expr = str_to_sympy(eq.expr.code, variables).xreplace(
                    substitutions
                )
                new_str_expr = sympy_to_str(new_sympy_expr)
                expr = Expression(new_str_expr)

                if eq.type == SUBEXPRESSION:
                    if eq.var_type == INTEGER:
                        sympy_var = sympy.Symbol(eq.varname, integer=True)
                    else:
                        sympy_var = sympy.Symbol(eq.varname, real=True)
                    substitutions.update(
                        {sympy_var: str_to_sympy(expr.code, variables)}
                    )
                    self._substituted_expressions.append((eq.varname, expr))
                elif eq.type == DIFFERENTIAL_EQUATION:
                    #  a differential equation that we have to check
                    self._substituted_expressions.append((eq.varname, expr))
                else:
                    raise AssertionError(f"Unknown equation type {eq.type}")

        if include_subexpressions:
            return self._substituted_expressions
        else:
            return [
                (name, expr)
                for name, expr in self._substituted_expressions
                if self[name].type == DIFFERENTIAL_EQUATION
            ]

    def _get_stochastic_type(self):
        """
        Returns the type of stochastic differential equations (additivive or
        multiplicative). The system is only classified as ``additive`` if *all*
        equations have only additive noise (or no noise).

        Returns
        -------
        type : str
            Either ``None`` (no noise variables), ``'additive'`` (factors for
            all noise variables are independent of other state variables or
            time), ``'multiplicative'`` (at least one of the noise factors
            depends on other state variables and/or time).
        """

        if not self.is_stochastic:
            return None

        for _, expr in self.get_substituted_expressions():
            _, stochastic = expr.split_stochastic()
            if stochastic is not None:
                for factor in stochastic.values():
                    if "t" in factor.identifiers:
                        # noise factor depends on time
                        return "multiplicative"

                    for identifier in factor.identifiers:
                        if identifier in self.diff_eq_names:
                            # factor depends on another state variable
                            return "multiplicative"

        return "additive"

    ############################################################################
    # Properties
    ############################################################################

    # Lists of equations or (variable, expression tuples)
    ordered = property(
        lambda self: sorted(
            self._equations.values(), key=lambda key: (key.update_order, key.varname)
        ),
        doc=(
            "A list of all equations, sorted "
            "according to the order in which they should "
            "be updated"
        ),
    )

    diff_eq_expressions = property(
        lambda self: [
            (varname, eq.expr)
            for varname, eq in self.items()
            if eq.type == DIFFERENTIAL_EQUATION
        ],
        doc=(
            "A list of (variable name, expression) "
            "tuples of all differential equations."
        ),
    )

    eq_expressions = property(
        lambda self: [
            (varname, eq.expr)
            for varname, eq in self.items()
            if eq.type in (SUBEXPRESSION, DIFFERENTIAL_EQUATION)
        ],
        doc="A list of (variable name, expression) tuples of all equations.",
    )

    # Sets of names

    names = property(
        lambda self: {eq.varname for eq in self.ordered},
        doc="All variable names defined in the equations.",
    )

    diff_eq_names = property(
        lambda self: {
            eq.varname for eq in self.ordered if eq.type == DIFFERENTIAL_EQUATION
        },
        doc="All differential equation names.",
    )

    subexpr_names = property(
        lambda self: {eq.varname for eq in self.ordered if eq.type == SUBEXPRESSION},
        doc="All subexpression names.",
    )

    eq_names = property(
        lambda self: {
            eq.varname
            for eq in self.ordered
            if eq.type in (DIFFERENTIAL_EQUATION, SUBEXPRESSION)
        },
        doc="All equation names (including subexpressions).",
    )

    parameter_names = property(
        lambda self: {eq.varname for eq in self.ordered if eq.type == PARAMETER},
        doc="All parameter names.",
    )

    dimensions = property(
        lambda self: {var: eq.dim for var, eq in self._equations.items()},
        doc=(
            "Dictionary of all internal variables and their "
            "corresponding physical dimensions."
        ),
    )

    identifiers = property(
        lambda self: set().union(*[eq.identifiers for eq in self._equations.values()])
        - self.names,
        doc=(
            "Set of all identifiers used in the equations, "
            "excluding the variables defined in the equations"
        ),
    )

    stochastic_variables = property(
        lambda self: {
            variable
            for variable in self.identifiers
            if variable == "xi" or variable.startswith("xi_")
        }
    )

    # general properties
    is_stochastic = property(
        lambda self: len(self.stochastic_variables) > 0,
        doc="Whether the equations are stochastic.",
    )

    stochastic_type = property(fget=_get_stochastic_type)

    def _sort_subexpressions(self):
        """
        Sorts the subexpressions in a way that resolves their dependencies
        upon each other. After this method has been run, the subexpressions
        returned by the ``ordered`` property are in the order in which
        they should be updated
        """

        # Get a dictionary of all the dependencies on other subexpressions,
        # i.e. ignore dependencies on parameters and differential equations
        static_deps = {}
        for eq in self._equations.values():
            if eq.type == SUBEXPRESSION:
                static_deps[eq.varname] = [
                    dep
                    for dep in eq.identifiers
                    if dep in self._equations
                    and self._equations[dep].type == SUBEXPRESSION
                ]

        try:
            sorted_eqs = topsort(static_deps)
        except ValueError:
            raise ValueError(
                "Cannot resolve dependencies between static "
                "equations, dependencies contain a cycle."
            )

        # put the equations objects in the correct order
        for order, static_variable in enumerate(sorted_eqs):
            self._equations[static_variable].update_order = order

        # Sort differential equations and parameters after subexpressions
        for eq in self._equations.values():
            if eq.type == DIFFERENTIAL_EQUATION:
                eq.update_order = len(sorted_eqs)
            elif eq.type == PARAMETER:
                eq.update_order = len(sorted_eqs) + 1

    @property
    def dependencies(self):
        """
        Calculate the dependencies of all differential equations and
        subexpressions.
        """
        # Create a dictionary mapping differential equations and
        # subexpressions to a list of their dependencies within the equations
        # (ignoring external constants, unit names, etc.)
        # Note that a differential equation such as "dv/dt = -v / tau" does not
        # mean that the variable "v" depends on itself. To make the distinction between
        # a variable and its derivative, we use the variable name + the prime symbol
        # in this dictionary.
        # As an example, the equations:
        #   dv/dt = I_m / C_m : volt
        #   I_m = I_ext + I_pas : amp
        #   I_ext = 1*nA + sin(2*pi*100*Hz*t)*nA : amp
        #   I_pas = g_L*(E_L - v) : amp
        # would be translated into the following dictionary
        #  {"v" : [],
        #   "v'": ["I_m"]
        #   "I_m": ["I_ext", "I_pas"],
        #   "I_ext": [],
        #   "I_pas": ["v"] }
        deps = {}
        for eq in self._equations.values():
            if eq.type == SUBEXPRESSION:
                name = eq.varname
            elif eq.type == DIFFERENTIAL_EQUATION:
                name = eq.varname + "'"
                deps[eq.varname] = []
            else:
                continue
            deps[name] = [
                dep
                for dep in eq.identifiers
                if dep in self._equations and self._equations[dep].type != PARAMETER
            ]
        try:
            sorted_eqs = topsort(deps)
        except ValueError:
            raise ValueError(
                "Cannot resolve dependencies between static "
                "equations, dependencies contain a cycle."
            )
        # Remove the dummy entries for differential equations and rename
        # x' → x
        sorted_eqs = [
            x.replace("'", "") for x in sorted_eqs if x not in self.diff_eq_names
        ]
        # Now recursively fill in the dependencies – this only needs a single
        # pass due to the previous sorting
        deps = {}
        Dependency = namedtuple(
            "Dependency", ["equation", "via"], defaults=((),)
        )  # default for via is empty tuple
        for eq in sorted_eqs:
            dep_names = {
                dep for dep in self._equations[eq].identifiers if dep in self._equations
            }
            deps[eq] = [Dependency(equation=self._equations[dep]) for dep in dep_names]
            # add all indirect dependencies
            for dep in dep_names:
                for indirect_dep in deps.get(dep, []):
                    if indirect_dep.equation.varname == dep:
                        continue  # do not go into recursion if a variable depends on itself
                    if any(
                        indirect_dep.equation.varname == existing_dep.equation.varname
                        for existing_dep in deps[eq]
                    ):
                        continue  # Do not add indirect dependencies for things we also depend on directly
                    deps[eq].append(
                        Dependency(
                            equation=indirect_dep.equation,
                            via=(dep,) + indirect_dep.via,
                        )
                    )
        return deps

    def check_units(self, group, run_namespace):
        """
        Check all the units for consistency.

        Parameters
        ----------
        group : `Group`
            The group providing the context
        run_namespace : dict-like, optional
            An additional namespace that is used for variable lookup (if not
            defined, the implicit namespace of local variables is used).
        level : int, optional
            How much further to go up in the stack to find the calling frame

        Raises
        ------
        DimensionMismatchError
            In case of any inconsistencies.
        """
        all_variables = dict(group.variables)
        external = frozenset().union(
            *[expr.identifiers for _, expr in self.eq_expressions]
        )
        external -= set(all_variables.keys())

        resolved_namespace = group.resolve_all(
            external, run_namespace, user_identifiers=external
        )  # all variables are user defined

        all_variables.update(resolved_namespace)
        for var, eq in self._equations.items():
            if eq.type == PARAMETER:
                # no need to check units for parameters
                continue

            if eq.type == DIFFERENTIAL_EQUATION:
                try:
                    check_dimensions(
                        str(eq.expr), self.dimensions[var] / second.dim, all_variables
                    )
                except DimensionMismatchError as ex:
                    raise DimensionMismatchError(
                        "Inconsistent units in "
                        "differential equation "
                        f"defining variable '{eq.varname}':"
                        f"\n{ex.desc}",
                        *ex.dims,
                    ) from ex
            elif eq.type == SUBEXPRESSION:
                try:
                    check_dimensions(str(eq.expr), self.dimensions[var], all_variables)
                except DimensionMismatchError as ex:
                    raise DimensionMismatchError(
                        "Inconsistent units in "
                        f"subexpression {eq.varname}:"
                        f"\n%{ex.desc}",
                        *ex.dims,
                    ) from ex
            else:
                raise AssertionError(f"Unknown equation type: '{eq.type}'")

    def check_flags(self, allowed_flags, incompatible_flags=None):
        """
        Check the list of flags.

        Parameters
        ----------
        allowed_flags : dict
             A dictionary mapping equation types (PARAMETER,
             DIFFERENTIAL_EQUATION, SUBEXPRESSION) to a list of strings (the
             allowed flags for that equation type)
        incompatible_flags : list of tuple
            A list of flag combinations that are not allowed for the same
            equation.
        Notes
        -----
        Not specifying allowed flags for an equation type is the same as
        specifying an empty list for it.

        Raises
        ------
        ValueError
            If any flags are used that are not allowed.
        """
        if incompatible_flags is None:
            incompatible_flags = []

        for eq in self.values():
            for flag in eq.flags:
                if eq.type not in allowed_flags or len(allowed_flags[eq.type]) == 0:
                    raise ValueError(
                        f"Equations of type '{eq.type}' cannot have any flags."
                    )
                if flag not in allowed_flags[eq.type]:
                    raise ValueError(
                        f"Equations of type '{eq.type}' cannot have a "
                        f"flag '{flag}', only the following flags "
                        f"are allowed: {allowed_flags[eq.type]}"
                    )
                # Check for incompatibilities
                for flag_combinations in incompatible_flags:
                    if flag in flag_combinations:
                        remaining_flags = set(flag_combinations) - {flag}
                        for remaining_flag in remaining_flags:
                            if remaining_flag in eq.flags:
                                raise ValueError(
                                    f"Flag '{flag}' cannot be "
                                    "combined with flag "
                                    f"'{remaining_flag}'"
                                )

    ############################################################################
    # Representation
    ############################################################################

    def __str__(self):
        strings = [str(eq) for eq in self.ordered]
        return "\n".join(strings)

    def __repr__(self):
        return f"<Equations object consisting of {len(self._equations)} equations>"

    def _latex(self, *args):
        equations = []
        for eq in self._equations.values():
            # do not use SingleEquations._latex here as we want nice alignment
            varname = sympy.Symbol(eq.varname)
            if eq.type == DIFFERENTIAL_EQUATION:
                lhs = r"\frac{\mathrm{d}" + sympy.latex(varname) + r"}{\mathrm{d}t}"
            else:
                # Normal equation or parameter
                lhs = varname
            if not eq.type == PARAMETER:
                rhs = str_to_sympy(eq.expr.code)
            if len(eq.flags):
                flag_str = ", flags: " + ", ".join(eq.flags)
            else:
                flag_str = ""
            if eq.type == PARAMETER:
                eq_latex = r"{} &&& \text{{(unit: ${}${})}}".format(
                    sympy.latex(lhs),
                    sympy.latex(get_unit(eq.dim)),
                    flag_str,
                )
            else:
                eq_latex = r"{} &= {} && \text{{(unit of ${}$: ${}${})}}".format(
                    lhs,  # already a string
                    sympy.latex(rhs),
                    sympy.latex(varname),
                    sympy.latex(get_unit(eq.dim)),
                    flag_str,
                )
            equations.append(eq_latex)
        return r"\begin{align*}" + (r"\\" + "\n").join(equations) + r"\end{align*}"

    def _repr_latex_(self):
        return sympy.latex(self)

    def _repr_pretty_(self, p, cycle):
        """Pretty printing for ipython"""
        if cycle:
            # Should never happen
            raise AssertionError("Cyclical call of 'Equations._repr_pretty_'")
        for eq in self._equations.values():
            p.pretty(eq)
            p.breakable("\n")


def is_stateful(expression, variables):
    """
    Whether the given expression refers to stateful functions (and is therefore
    not guaranteed to give the same result if called repetively).

    Parameters
    ----------
    expression : `sympy.Expression`
        The sympy expression to check.
    variables : dict
        The dictionary mapping variable names to `Variable` or `Function`
        objects.

    Returns
    -------
    stateful : bool
        ``True``, if the given expression refers to a stateful function like
        ``rand()`` and ``False`` otherwise.
    """
    func_name = str(expression.func)
    func_variable = variables.get(func_name, None)
    if func_variable is not None and not func_variable.stateless:
        return True
    for arg in expression.args:
        if is_stateful(arg, variables):
            return True
    return False


def check_subexpressions(group, equations, run_namespace):
    """
    Checks the subexpressions in the equations and raises an error if a
    subexpression refers to stateful functions without being marked as
    "constant over dt".

    Parameters
    ----------
    group : `Group`
        The group providing the context.
    equations : `Equations`
        The equations to check.
    run_namespace : dict
        The run namespace for resolving variables.

    Raises
    ------
    SyntaxError
        For subexpressions not marked as "constant over dt" that refer to
        stateful functions.
    """
    for eq in equations.ordered:
        if eq.type == SUBEXPRESSION:
            # Check whether the expression is stateful (most commonly by
            # referring to rand() or randn()
            variables = group.resolve_all(
                eq.identifiers,
                run_namespace,
                # we don't need to raise any warnings
                # for the user here, warnings will
                # be raised in create_runner_codeobj
                user_identifiers=set(),
            )
            expression = str_to_sympy(eq.expr.code, variables=variables)

            # Check whether the expression refers to stateful functions
            if is_stateful(expression, variables):
                raise SyntaxError(
                    f"The subexpression '{eq.varname}' refers to a "
                    "stateful function (e.g. rand()). Such "
                    "expressions should only be evaluated "
                    "once per timestep, add the 'constant "
                    "over dt' flag."
                )


def extract_constant_subexpressions(eqs):
    without_const_subexpressions = []
    const_subexpressions = []
    for eq in eqs.ordered:
        if eq.type == SUBEXPRESSION and "constant over dt" in eq.flags:
            flags = set(eq.flags) - {"constant over dt"}
            without_const_subexpressions.append(
                SingleEquation(
                    PARAMETER, eq.varname, eq.dim, var_type=eq.var_type, flags=flags
                )
            )
            const_subexpressions.append(eq)
        else:
            without_const_subexpressions.append(eq)

    return (Equations(without_const_subexpressions), Equations(const_subexpressions))



#!/usr/bin/env python3
"""
Fig. 3 of

Simple Model of Spiking Neurons
IEEE Transactions on Neural Networks ( Volume: 14, Issue: 6, Nov. 2003)
Eugene M. Izhikevich

based on net.m by Eugene M. Izhikevich (http://izhikevich.org/publications/spikes.htm)

Akif Erdem Sağtekin and Sebastian Schmitt, 2022
"""
import matplotlib.pyplot as plt
import numpy as np

from brian2 import NeuronGroup, Synapses, SpikeMonitor, StateMonitor
from brian2 import ms, mV
from brian2 import defaultclock, run

tfinal = 1000 * ms
Ne = 800
Ni = 200

re = np.random.uniform(size=Ne)
ri = np.random.uniform(size=Ni)
weights = np.hstack(
    [
        0.5 * np.random.uniform(size=(Ne + Ni, Ne)),
        -np.random.uniform(size=(Ne + Ni, Ni)),
    ]
).T

defaultclock.dt = 1 * ms

eqs = """dv/dt = (0.04*v**2 + 5*v + 140 - u + I + I_noise )/ms : 1
         du/dt = (a*(b*v - u))/ms  : 1
         I : 1
         I_noise : 1
         a : 1
         b : 1
         c : 1
         d : 1
       """

N = NeuronGroup(Ne + Ni, eqs, threshold="v>=30", reset="v=c; u+=d", method="euler")
N.v = -65

N_exc = N[:Ne]
N_inh = N[Ne:]

spikemon = SpikeMonitor(N)
statemon = StateMonitor(N, 'v', record=0, when='after_thresholds')
N_exc.a = 0.02
N_exc.b = 0.2
N_exc.c = -65 + 15 * re**2
N_exc.d = 8 - 6 * re**2

N_inh.a = 0.02 + 0.08 * ri
N_inh.b = 0.25 - 0.05 * ri
N_inh.c = -65
N_inh.d = 2

N_exc.u = "b*v"
N_inh.u = "b*v"

S = Synapses(
    N,
    N,
    "w : 1",
    on_pre={"up": "I += w", "down": "I -= w"},
    delay={"up": 0 * ms, "down": 1 * ms},
)
S.connect()
S.w[:] = weights.flatten()

N_exc.run_regularly("I_noise = 5*randn()", dt=1 * ms)
N_inh.run_regularly("I_noise = 2*randn()", dt=1 * ms)

run(tfinal)

fig, (ax, ax_voltage) = plt.subplots(2, 1, sharex=True,
                                     gridspec_kw={'height_ratios': (3, 1)})

ax.scatter(spikemon.t / ms, spikemon.i[:], marker="_", color="k", s=10)
ax.set_xlim(0, tfinal / ms)
ax.set_ylim(0, len(N))
ax.set_ylabel("neuron number")
ax.set_yticks(np.arange(0, len(N), 100))
ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
ax.axhline(Ne, color="k")
ax.text(500, 900, 'inhibitory', backgroundcolor='w', color='k', ha='center')
ax.text(500, 400, 'excitatory', backgroundcolor='w', color='k', ha='center')

ax_voltage.plot(statemon.t / ms, np.clip(statemon.v[0], -np.inf, 30),
               color='k')
ax_voltage.text(25, 0, 'v₁(t)')
ax_voltage.set_xticks(np.arange(0, tfinal / ms, 100))
ax_voltage.spines['right'].set_visible(False)
ax_voltage.spines['top'].set_visible(False)
ax_voltage.set_xlabel("time, ms")

plt.show()



'''
Script to run the test suite during automatic testing (easier than putting all
the logic into Windows batch/bash statements)
'''
# Importing multiprocessing here seems to fix hangs in the test suite on OS X
# see https://github.com/scipy/scipy/issues/11835
import multiprocessing
# Prevent potential issues on multi-threaded execution
multiprocessing.set_start_method('spawn', force=True)
import os
import sys

import numpy as np

import brian2

if __name__ == '__main__':
    split_run = os.environ.get('SPLIT_RUN', None)
    standalone = os.environ.get('STANDALONE', 'no').lower() in ['yes', 'true']
    python_version = os.environ.get('PYTHON_VERSION', os.environ.get('PYTHON'))
    # If TRAVIS_OS_NAME is not defined, we are testing on appveyor
    operating_system = os.environ.get('AGENT_OS', 'unknown').lower()
    cross_compiled = os.environ.get('CROSS_COMPILED', 'FALSE').lower() in ['yes', 'true']
    do_not_reset_preferences = os.environ.get('DO_NOT_RESET_PREFERENCES', 'false').lower() in ['yes', 'true']
    dtype_32_bit = os.environ.get('FLOAT_DTYPE_32', 'no').lower() in ['yes', 'true']
    sphinx_dir = os.environ.get('SPHINX_DIR')
    src_dir = os.environ.get('SRCDIR')
    deprecation_error = os.environ.get('DEPRECATION_ERROR', 'false').lower() in ['yes', 'true']
    if split_run == '1':
        targets = ['numpy']
        independent = True
    elif split_run == '2':
        targets = ['cython']
        independent = False
    else:
        targets = None
        independent = True

    if operating_system == 'windows' or standalone:
        in_parallel = []
    else:
        in_parallel = ['codegen_independent', 'numpy', 'cpp_standalone']

    if operating_system in ['linux', 'windows_nt']:
        openmp = True
    else:
        openmp = False

    reset_preferences = not (cross_compiled or do_not_reset_preferences)
    if dtype_32_bit:
        float_dtype = np.float32
    else:
        float_dtype = None

    if deprecation_error:
        args = ['-W', 'error::DeprecationWarning', '--tb=short']
    else:
        # Use coverage when running on GitHub
        if "GITHUB_WORKSPACE" in os.environ:
            args = [
                "--cov",
                "--cov-append",
                "--cov-report",
                "xml",
                "--cov-report",
                "term",
                "--cov-config",
                os.path.join(os.environ["GITHUB_WORKSPACE"], ".coveragerc"),
            ]

    if standalone:
        result = brian2.test([],
                             test_codegen_independent=False,
                             test_standalone='cpp_standalone',
                             test_openmp=openmp,
                             test_in_parallel=in_parallel,
                             reset_preferences=reset_preferences,
                             float_dtype=float_dtype,
                             test_GSL=True,
                             sphinx_dir=sphinx_dir,
                             additional_args=args)
    else:
        result = brian2.test(targets,
                             test_codegen_independent=independent,
                             test_standalone=None,
                             test_in_parallel=in_parallel,
                             reset_preferences=reset_preferences,
                             float_dtype=float_dtype,
                             test_GSL=True,
                             sphinx_dir=sphinx_dir,
                             additional_args=args)

    if not result:
        sys.exit(1)



'''
Brette R (2013). Sharpness of spike initiation in neurons explained by compartmentalization.
PLoS Comp Biol, doi: 10.1371/journal.pcbi.1003338.

Fig. 5A. Voltage trace for current injection, with an additional reset when a spike is produced.

Trick: to reset the entire neuron, we use a set of synapses from the spike initiation compartment where the
threshold condition applies to all compartments, and the reset operation (v = EL) is applied there every time
a spike is produced.
'''
from brian2 import *
from params import *

defaultclock.dt = 0.025*ms
duration = 500*ms

# Morphology
morpho = Soma(50*um)  # chosen for a target Rm
morpho.axon = Cylinder(diameter=1*um, length=300*um, n=300)

# Input
taux = 5*ms
sigmax = 12*mV
xx0 = 7*mV

compartment = 40

# Channels
eqs = '''
Im = gL * (EL - v) + gNa * m * (ENa - v) + gLx * (xx0 + xx) : amp/meter**2
dm/dt = (minf - m) / taum : 1  # simplified Na channel
minf = 1 / (1 + exp((va - v) / ka)) : 1
gNa : siemens/meter**2
gLx : siemens/meter**2
dxx/dt = -xx / taux + sigmax * (2 / taux)**.5 *xi : volt
'''

neuron = SpatialNeuron(morphology=morpho, model=eqs, Cm=Cm, Ri=Ri,
                       threshold='m>0.5', threshold_location=compartment,
                       refractory=5*ms)
neuron.v = EL
neuron.gLx[0] = gL
neuron.gNa[compartment] = gNa_0 / neuron.area[compartment]

# Reset the entire neuron when there is a spike
reset = Synapses(neuron, neuron, on_pre='v = EL')
reset.connect('i == compartment')  # Connects the spike initiation compartment to all compartments

# Monitors
S = SpikeMonitor(neuron)
M = StateMonitor(neuron, 'v', record=0)
run(duration, report='text')

# Add spikes for display
v = M[0].v
for t in S.t:
    v[int(t / defaultclock.dt)] = 50*mV

plot(M.t/ms, v/mV, 'k')
tight_layout()
show()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
This is an implementation of a benchmark described
in the following review paper:

Simulation of networks of spiking neurons: A review of tools and strategies (2006).
Brette, Rudolph, Carnevale, Hines, Beeman, Bower, Diesmann, Goodman, Harris, Zirpe,
Natschläger, Pecevski, Ermentrout, Djurfeldt, Lansner, Rochel, Vibert, Alvarez, Muller,
Davison, El Boustani and Destexhe.
Journal of Computational Neuroscience

Benchmark 3: random network of HH neurons with exponential synaptic conductances

Clock-driven implementation
(no spike time interpolation)

R. Brette - Dec 2007
"""

from brian2 import *

# Parameters
area = 20000*umetre**2
Cm = (1*ufarad*cm**-2) * area
gl = (5e-5*siemens*cm**-2) * area

El = -60*mV
EK = -90*mV
ENa = 50*mV
g_na = (100*msiemens*cm**-2) * area
g_kd = (30*msiemens*cm**-2) * area
VT = -63*mV
# Time constants
taue = 5*ms
taui = 10*ms
# Reversal potentials
Ee = 0*mV
Ei = -80*mV
we = 6*nS  # excitatory synaptic weight
wi = 67*nS  # inhibitory synaptic weight

# The model
eqs = Equations('''
dv/dt = (gl*(El-v)+ge*(Ee-v)+gi*(Ei-v)-
         g_na*(m*m*m)*h*(v-ENa)-
         g_kd*(n*n*n*n)*(v-EK))/Cm : volt
dm/dt = alpha_m*(1-m)-beta_m*m : 1
dn/dt = alpha_n*(1-n)-beta_n*n : 1
dh/dt = alpha_h*(1-h)-beta_h*h : 1
dge/dt = -ge*(1./taue) : siemens
dgi/dt = -gi*(1./taui) : siemens
alpha_m = 0.32*(mV**-1)*4*mV/exprel((13*mV-v+VT)/(4*mV))/ms : Hz
beta_m = 0.28*(mV**-1)*5*mV/exprel((v-VT-40*mV)/(5*mV))/ms : Hz
alpha_h = 0.128*exp((17*mV-v+VT)/(18*mV))/ms : Hz
beta_h = 4./(1+exp((40*mV-v+VT)/(5*mV)))/ms : Hz
alpha_n = 0.032*(mV**-1)*5*mV/exprel((15*mV-v+VT)/(5*mV))/ms : Hz
beta_n = .5*exp((10*mV-v+VT)/(40*mV))/ms : Hz
''')

P = NeuronGroup(4000, model=eqs, threshold='v>-20*mV', refractory=3*ms,
                method='exponential_euler')
Pe = P[:3200]
Pi = P[3200:]
Ce = Synapses(Pe, P, on_pre='ge+=we')
Ci = Synapses(Pi, P, on_pre='gi+=wi')
Ce.connect(p=0.02)
Ci.connect(p=0.02)

# Initialization
P.v = 'El + (randn() * 5 - 5)*mV'
P.ge = '(randn() * 1.5 + 4) * 10.*nS'
P.gi = '(randn() * 12 + 20) * 10.*nS'

# Record a few traces
trace = StateMonitor(P, 'v', record=[1, 10, 100])
run(1 * second, report='text')
plot(trace.t/ms, trace[1].v/mV)
plot(trace.t/ms, trace[10].v/mV)
plot(trace.t/ms, trace[100].v/mV)
xlabel('t (ms)')
ylabel('v (mV)')
show()



"""
Check that the features of `NeuronGroup` are available and correct.
"""

from brian2 import *
from brian2.tests.features import FeatureTest, InaccuracyError


class NeuronGroupIntegrationLinear(FeatureTest):
    category = "NeuronGroup"
    name = "Linear integration"
    tags = ["NeuronGroup", "run"]

    def run(self):
        self.tau = tau = 1 * second
        self.v_init = linspace(0.1, 1, 10)
        self.duration = 100 * ms
        G = self.G = NeuronGroup(10, "dv/dt=-v/tau:1")
        self.G.v = self.v_init
        run(self.duration)

    def results(self):
        v_correct = self.v_init * exp(-self.duration / self.tau)
        return v_correct, self.G.v[:]

    def check(self, maxrelerr, correct_end):
        v_correct, v_end = correct_end
        err = amax(abs(v_end - v_correct) / v_correct)
        if err > maxrelerr:
            raise InaccuracyError(err)


class NeuronGroupIntegrationEuler(FeatureTest):
    category = "NeuronGroup"
    name = "Euler integration"
    tags = ["NeuronGroup", "run"]

    def run(self):
        self.tau = tau = 1 * second
        self.v_init = linspace(0.1, 1, 10)
        self.duration = 100 * ms
        G = self.G = NeuronGroup(10, "dv/dt=-v**1.1/tau:1")
        self.G.v = self.v_init
        run(self.duration)

    def results(self):
        return self.G.v[:]

    compare = FeatureTest.compare_arrays


class NeuronGroupLIF(FeatureTest):
    category = "NeuronGroup"
    name = "Leaky integrate and fire"
    tags = ["NeuronGroup", "Threshold", "Reset", "run", "SpikeMonitor"]

    def run(self):
        self.tau = tau = 10 * ms
        self.duration = 1000 * ms
        G = self.G = NeuronGroup(1, "dv/dt=(1.2-v)/tau:1", threshold="v>1", reset="v=0")
        M = self.M = SpikeMonitor(self.G)
        run(self.duration)

    def results(self):
        return self.M.t[:]

    compare = FeatureTest.compare_arrays


class NeuronGroupLIFRefractory(FeatureTest):
    category = "NeuronGroup"
    name = "Refractory leaky integrate and fire"
    tags = ["NeuronGroup", "Threshold", "Reset", "Refractory", "run", "SpikeMonitor"]

    def run(self):
        self.tau = tau = 10 * ms
        self.duration = 1000 * ms
        G = self.G = NeuronGroup(
            1,
            "dv/dt=(1.2-v)/tau:1 (unless refractory)",
            threshold="v>1",
            reset="v=0",
            refractory=1 * ms,
        )
        M = self.M = SpikeMonitor(self.G)
        run(self.duration)

    def results(self):
        return self.M.t[:]

    compare = FeatureTest.compare_arrays


if __name__ == "__main__":
    #    ft = NeuronGroupIntegrationLinear()
    #    ft.run()
    #    res = ft.results()
    #    ft.check(res)

    ft = NeuronGroupIntegrationEuler()
    ft.run()
    res = ft.results()



import gc
import inspect
import itertools
import weakref

from brian2.units.allunits import second
from brian2.units.fundamentalunits import check_units
from brian2.utils.logger import get_logger

from .base import BrianObject
from .network import Network

__all__ = [
    "MagicNetwork",
    "magic_network",
    "MagicError",
    "run",
    "stop",
    "collect",
    "store",
    "restore",
    "start_scope",
]

logger = get_logger(__name__)


def _get_contained_objects(obj):
    """
    Helper function to recursively get all contained objects.

    Parameters
    ----------
    obj : `BrianObject`
        An object that (potentially) contains other objects, e.g. a
        `NeuronGroup` contains a `StateUpdater`, etc.

    Returns
    -------
    objects : list of `BrianObject`
        A list of all the objects contained in `obj`
    """
    objects = []
    contained_objects = getattr(obj, "contained_objects", [])
    objects.extend(contained_objects)
    for contained_obj in contained_objects:
        objects.extend(_get_contained_objects(contained_obj))

    return objects


def get_objects_in_namespace(level):
    r"""
    Get all the objects in the current namespace that derive from `BrianObject`.
    Used to determine the objects for the `MagicNetwork`.

    Parameters
    ----------
    level : int, optional
        How far to go back to get the locals/globals. Each function/method
        call should add ``1`` to this argument, functions/method with a
        decorator have to add ``2``.

    Returns
    -------
    objects : set
        A set with weak references to the `BrianObject`\ s in the namespace.
    """
    # Get the locals and globals from the stack frame
    objects = set()
    frame = inspect.stack()[level + 1][0]
    for _, v in itertools.chain(frame.f_globals.items(), frame.f_locals.items()):
        # We are only interested in numbers and functions, not in
        # everything else (classes, modules, etc.)
        if isinstance(v, BrianObject):
            objects.add(weakref.ref(v))
    del frame
    return objects


class MagicError(Exception):
    """
    Error that is raised when something goes wrong in `MagicNetwork`

    See notes to `MagicNetwork` for more details.
    """

    pass


class MagicNetwork(Network):
    """
    `Network` that automatically adds all Brian objects

    In order to avoid bugs, this class will occasionally raise
    `MagicError` when the intent of the user is not clear. See the notes
    below for more details on this point. If you persistently see this
    error, then Brian is not able to safely guess what you intend to do, and
    you should use a `Network` object and call `Network.run` explicitly.

    Note that this class cannot be instantiated by the user, there can be only
    one instance `magic_network` of `MagicNetwork`.

    Notes
    -----

    All Brian objects that are visible at the point of the `run` call will be
    included in the network. This class is designed to work in the following
    two major use cases:

    1. You create a collection of Brian objects, and call `run` to run the
       simulation. Subsequently, you may call `run` again to run it again for
       a further duration. In this case, the `Network.t` time will start at 0
       and for the second call to `run` will continue from the end of the
       previous run.

    2. You have a loop in which at each iteration, you create some Brian
       objects and run a simulation using them. In this case, time is reset to
       0 for each call to `run`.

    In any other case, you will have to explicitly create a `Network` object
    yourself and call `Network.run` on this object. Brian has a built in
    system to guess which of the cases above applies and behave correctly.
    When it is not possible to safely guess which case you are in, it raises
    `MagicError`. The rules for this guessing system are explained below.

    If a simulation consists only of objects that have not been run, it will
    assume that you want to start a new simulation. If a simulation only
    consists of objects that have been simulated in the previous `run` call,
    it will continue that simulation at the previous time.

    If neither of these two situations apply, i.e., the network consists of a
    mix of previously run objects and new objects, an error will be raised.

    In these checks, "non-invalidating" objects (i.e. objects that have
    `BrianObject.invalidates_magic_network` set to ``False``) are ignored, e.g.
    creating new monitors is always possible.

    See Also
    --------
    Network, collect, run, stop, store, restore
    """

    _already_created = False

    def __init__(self):
        if MagicNetwork._already_created:
            raise ValueError("There can be only one MagicNetwork.")
        MagicNetwork._already_created = True

        super().__init__(name="magicnetwork*")

        self._previous_refs = set()

    def add(self, *objs):
        """
        You cannot add objects directly to `MagicNetwork`
        """
        raise MagicError("Cannot directly modify MagicNetwork")

    def remove(self, *objs):
        """
        You cannot remove objects directly from `MagicNetwork`
        """
        raise MagicError("Cannot directly modify MagicNetwork")

    def _update_magic_objects(self, level):
        objects = collect(level + 1)
        contained_objects = set()
        for obj in objects:
            for contained in _get_contained_objects(obj):
                contained_objects.add(contained)
        objects |= contained_objects

        # check whether we should restart time, continue time, or raise an
        # error
        some_known = False
        some_new = False
        for obj in objects:
            if obj._network == self.id:
                some_known = True  # we are continuing a previous run
            elif obj._network is None and obj.invalidates_magic_network:
                some_new = True
            # Note that the inclusion of objects that have been run as part of
            # other objects will lead to an error in `Network.before_run`, we
            # do not have to deal with this case here.

        if some_known and some_new:
            raise MagicError(
                "The magic network contains a mix of objects "
                "that has been run before and new objects, Brian "
                "does not know whether you want to start a new "
                "simulation or continue an old one. Consider "
                "explicitly creating a Network object. Also note "
                "that you can find out which objects will be "
                "included in a magic network with the "
                "collect() function."
            )
        elif some_new:  # all objects are new, start a new simulation
            # reset time
            self.t_ = 0.0
            # reset id -- think of this as a new Network
            self.assign_id()

        for obj in objects:
            if obj._network is None:
                obj._network = self.id

        self.objects = objects
        numobjs = len(self.objects)
        names = ", ".join(obj.name for obj in self.objects)
        logger.debug(
            f"Updated MagicNetwork to include {numobjs} objects with names {names}",
            name_suffix="magic_objects",
        )

    def check_dependencies(self):
        all_ids = {obj.id for obj in self.objects}
        for obj in self.objects:
            if not obj.active:
                continue  # object is already inactive, no need to check it
            for dependency in obj._dependencies:
                if dependency not in all_ids:
                    logger.warn(
                        f"'{obj.name}' has been included in the network but "
                        "not the object on which it depends."
                        f"Setting '{obj.name}' to inactive.",
                        name_suffix="dependency_warning",
                    )
                    obj.active = False
                    break

    def after_run(self):
        super().after_run()
        self.objects.clear()
        gc.collect()  # Make sure that all unused objects are cleared

    def run(
        self,
        duration,
        report=None,
        report_period=10 * second,
        namespace=None,
        profile=None,
        level=0,
    ):
        self._update_magic_objects(level=level + 1)
        Network.run(
            self,
            duration,
            report=report,
            report_period=report_period,
            namespace=namespace,
            profile=profile,
            level=level + 1,
        )

    def store(self, name="default", filename=None, level=0):
        """
        See `Network.store`.
        """
        self._update_magic_objects(level=level + 1)
        super().store(name=name, filename=filename)
        self.objects.clear()

    def restore(
        self, name="default", filename=None, restore_random_state=False, level=0
    ):
        """
        See `Network.restore`.
        """
        self._update_magic_objects(level=level + 1)
        super().restore(
            name=name, filename=filename, restore_random_state=restore_random_state
        )
        self.objects.clear()

    def get_states(self, units=True, format="dict", subexpressions=False, level=0):
        """
        See `Network.get_states`.
        """
        self._update_magic_objects(level=level + 1)
        states = super().get_states(units, format, subexpressions, level=level + 1)
        self.objects.clear()
        return states

    def set_states(self, values, units=True, format="dict", level=0):
        """
        See `Network.set_states`.
        """
        self._update_magic_objects(level=level + 1)
        super().set_states(values, units, format, level=level + 1)
        self.objects.clear()

    def __str__(self):
        return "MagicNetwork()"

    __repr__ = __str__


#: Automatically constructed `MagicNetwork` of all Brian objects
magic_network = MagicNetwork()


def collect(level=0):
    r"""
    Return the list of `BrianObject`\ s that will be simulated if `run` is
    called.

    Parameters
    ----------
    level : int, optional
        How much further up to go in the stack to find the objects. Needs
        only to be specified if `collect` is called as part of a function
        and should be increased by 1 for every level of nesting. Defaults to 0.

    Returns
    -------
    objects : set of `BrianObject`
        The objects that will be simulated.
    """
    all_objects = set()
    for obj in get_objects_in_namespace(level=level + 1):
        obj = obj()
        if obj.add_to_magic_network:
            gk = BrianObject._scope_current_key
            k = obj._scope_key
            if gk != k:
                continue
            all_objects.add(obj)
    return all_objects


@check_units(duration=second, report_period=second)
def run(
    duration,
    report=None,
    report_period=10 * second,
    namespace=None,
    profile=None,
    level=0,
):
    """
    run(duration, report=None, report_period=10*second, namespace=None, level=0)

    Runs a simulation with all "visible" Brian objects for the given duration.
    Calls `collect` to gather all the objects, the simulation can
    be stopped by calling the global `stop` function.

    In order to avoid bugs, this function will occasionally raise
    `MagicError` when the intent of the user is not clear. See the notes to
    `MagicNetwork` for more details on this point. If you persistently see this
    error, then Brian is not able to safely guess what you intend to do, and
    you should use a `Network` object and call `Network.run` explicitly.

    Parameters
    ----------

    duration : `Quantity`
        The amount of simulation time to run for. If the network consists of
        new objects since the last time `run` was called, the start time will
        be reset to 0. If `run` is called twice or more without changing the
        set of objects, the second and subsequent runs will start from the
        end time of the previous run. To explicitly reset the time to 0,
        do ``magic_network.t = 0*second``.
    report : {None, 'text', 'stdout', 'stderr', function}, optional
        How to report the progress of the simulation. If ``None``, do not
        report progress. If ``'text'`` or ``'stdout'`` is specified, print
        the progress to stdout. If ``'stderr'`` is specified, print the
        progress to stderr. Alternatively, you can specify a callback
        ``callable(elapsed, completed, start, duration)`` which will be passed
        the amount of time elapsed as a `Quantity`, the fraction ``completed``
        from 0.0 to 1.0, the ``start`` time of the simulation as a `Quantity`
        and the total duration of the simulation (in biological time)
        as a `Quantity`.
        The function will always be called at the beginning and the end
        (i.e. for fractions 0.0 and 1.0), regardless of the ``report_period``.
    report_period : `Quantity`
        How frequently (in real time) to report progress.
    profile : bool, optional
        Whether to record profiling information (see `Network.profiling_info`).
        Defaults to ``None`` (which will use the value set by ``set_device``,
        if any).
    namespace : dict-like, optional
        A namespace in which objects which do not define their own
        namespace will be run. If not namespace is given, the locals and
        globals around the run function will be used.
    level : int, optional
        How deep to go down the stack frame to look for the locals/global
        (see `namespace` argument). Only necessary under particular
        circumstances, e.g. when calling the run function as part of a
        function call or lambda expression. This is used in tests, e.g.:
        ``assert_raises(MagicError, lambda: run(1*ms, level=3))``.

    See Also
    --------

    Network.run, MagicNetwork, collect, start_scope, stop

    Raises
    ------

    MagicError
        Error raised when it was not possible for Brian to safely guess the
        intended use. See `MagicNetwork` for more details.
    """
    return magic_network.run(
        duration,
        report=report,
        report_period=report_period,
        namespace=namespace,
        profile=profile,
        level=2 + level,
    )


run.__module__ = __name__


def store(name="default", filename=None):
    """
    Store the state of the network and all included objects.

    Parameters
    ----------
    name : str, optional
        A name for the snapshot, if not specified uses ``'default'``.
    filename : str, optional
        A filename where the state should be stored. If not specified, the
        state will be stored in memory.

    See Also
    --------
    Network.store
    """
    magic_network.store(name=name, filename=filename, level=1)


def restore(name="default", filename=None, restore_random_state=False):
    """
    Restore the state of the network and all included objects.

    Parameters
    ----------
    name : str, optional
        The name of the snapshot to restore, if not specified uses
        ``'default'``.
    filename : str, optional
        The name of the file from where the state should be restored. If
        not specified, it is expected that the state exist in memory
        (i.e. `Network.store` was previously called without the ``filename``
        argument).
    restore_random_state : bool, optional
        Whether to restore the state of the random number generator. If set
        to ``True``, going back to an earlier state of the simulation will
        continue exactly where it left off, even if the simulation is
        stochastic. If set to ``False`` (the default), random numbers are
        independent between runs (except for explicitly set random seeds),
        regardless of whether `store`/`restore` has been used or not. Note
        that this also restores numpy's random number generator (since it is
        used internally by Brian), but it does *not* restore Python's
        builtin random number generator in the ``random`` module.
    See Also
    --------
    Network.restore
    """
    magic_network.restore(
        name=name, filename=filename, restore_random_state=restore_random_state, level=1
    )


def stop():
    """
    Stops all running simulations.

    See Also
    --------

    Network.stop, run, reinit
    """
    Network._globally_stopped = True


def start_scope():
    """
    Starts a new scope for magic functions

    All objects created before this call will no longer be automatically
    included by the magic functions such as `run`.
    """
    BrianObject._scope_current_key += 1



"""
Module to support caching of function results to memory (used to cache results
of parsing, generation of state update code, etc.). Provides the `cached`
decorator.
"""

import collections
import functools
from collections.abc import Mapping


class CacheKey:
    """
    Mixin class for objects that will be used as keys for caching (e.g.
    `Variable` objects) and have to define a certain "identity" with respect
    to caching. This "identity" is different from standard Python hashing and
    equality checking: a `Variable` for example would be considered "identical"
    for caching purposes regardless which object (e.g. `NeuronGroup`) it belongs
    to (because this does not matter for parsing, creating abstract code, etc.)
    but this of course matters for the values it refers to and therefore for
    comparison of equality to other variables.

    Classes that mix in the `CacheKey` class should re-define the
    ``_cache_irrelevant_attributes`` attribute to note all the attributes that
    should be ignored. The property ``_state_tuple`` will refer to a tuple of
    all attributes that were not excluded in such a way; this tuple will be used
    as the key for caching purposes.
    """

    #: Set of attributes that should not be considered for caching of state
    #: update code, etc.
    _cache_irrelevant_attributes = set()

    @property
    def _state_tuple(self):
        """A tuple with this object's attribute values, defining its identity
        for caching purposes. See `CacheKey` for details."""
        return tuple(
            value
            for key, value in sorted(self.__dict__.items())
            if key not in self._cache_irrelevant_attributes
        )


class _CacheStatistics:
    """
    Helper class to store cache statistics
    """

    def __init__(self):
        self.hits = 0
        self.misses = 0

    def __repr__(self):
        return f"<Cache statistics: {int(self.hits)} hits, {int(self.misses)} misses>"


def cached(func):
    """
    Decorator to cache a function so that it will not be re-evaluated when
    called with the same arguments. Uses the `_hashable` function to make
    arguments usable as a dictionary key even though they mutable (lists,
    dictionaries, etc.).

    Notes
    -----
    This is *not* a general-purpose caching decorator in any way comparable to
    ``functools.lru_cache`` or joblib's caching functions. It is very simplistic
    (no maximum cache size, no normalization of calls, e.g. ``foo(3)`` and
    ``foo(x=3)`` are not considered equivalent function calls) and makes very
    specific assumptions for our use case. Most importantly, `Variable` objects
    are considered to be identical when they refer to the same object, even
    though the actually stored values might have changed.

    Parameters
    ----------
    func : function
        The function to decorate.

    Returns
    -------
    decorated : function
        The decorated function.
    """
    # For simplicity, we store the cache in the function itself
    func._cache = {}
    func._cache_statistics = _CacheStatistics()

    @functools.wraps(func)
    def cached_func(*args, **kwds):
        try:
            cache_key = tuple(
                [_hashable(arg) for arg in args]
                + [(key, _hashable(value)) for key, value in sorted(kwds.items())]
            )
        except TypeError:
            # If we cannot handle a type here, that most likely means that the
            # user provided an argument of a type we don't handle. This will
            # lead to an error message later that is most likely more meaningful
            # to the user than an error message by the caching system
            # complaining about an unsupported type.
            return func(*args, **kwds)
        if cache_key in func._cache:
            func._cache_statistics.hits += 1
        else:
            func._cache_statistics.misses += 1
            func._cache[cache_key] = func(*args, **kwds)
        return func._cache[cache_key]

    return cached_func


_of_type_cache = collections.defaultdict(set)


def _of_type(obj_type, check_type):
    if (obj_type, check_type) not in _of_type_cache:
        _of_type_cache[(obj_type, check_type)] = issubclass(obj_type, check_type)
    return _of_type_cache[(obj_type, check_type)]


def _hashable(obj):
    """Helper function to make a few data structures hashable (e.g. a
    dictionary gets converted to a frozenset). The function is specifically
    tailored to our use case and not meant to be generally useful."""
    if hasattr(obj, "_state_tuple"):
        return _hashable(obj._state_tuple)
    obj_type = type(obj)
    if _of_type(obj_type, Mapping):
        return frozenset(
            (_hashable(key), _hashable(value)) for key, value in obj.items()
        )
    elif _of_type(obj_type, set):
        return frozenset(_hashable(el) for el in obj)
    elif _of_type(obj_type, tuple) or _of_type(obj_type, list):
        return tuple(_hashable(el) for el in obj)
    if hasattr(obj, "dim") and getattr(obj, "shape", None) == ():
        # Scalar Quantity object
        return float(obj), obj.dim
    else:
        try:
            # Make sure that the object is hashable
            hash(obj)
            return obj
        except TypeError:
            raise TypeError(f"Do not know how to handle object of type {type(obj)}")



#!/usr/bin/env python
"""
Adapted from
Theory of Arachnid Prey Localization
W. Sturzl, R. Kempter, and J. L. van Hemmen
PRL 2000

Poisson inputs are replaced by integrate-and-fire neurons

Romain Brette
"""
from brian2 import *

# Parameters
degree = 2 * pi / 360.
duration = 500*ms
R = 2.5*cm  # radius of scorpion
vr = 50*meter/second  # Rayleigh wave speed
phi = 144*degree  # angle of prey
A = 250*Hz
deltaI = .7*ms  # inhibitory delay
gamma = (22.5 + 45 * arange(8)) * degree  # leg angle
delay = R / vr * (1 - cos(phi - gamma))   # wave delay

# Wave (vector w)
time = arange(int(duration / defaultclock.dt) + 1) * defaultclock.dt
Dtot = 0.
w = 0.
for f in arange(150, 451)*Hz:
    D = exp(-(f/Hz - 300) ** 2 / (2 * (50 ** 2)))
    rand_angle = 2 * pi * rand()
    w += 100 * D * cos(2 * pi * f * time + rand_angle)
    Dtot += D
w = .01 * w / Dtot

# Rates from the wave
rates = TimedArray(w, dt=defaultclock.dt)

# Leg mechanical receptors
tau_legs = 1 * ms
sigma = .01
eqs_legs = """
dv/dt = (1 + rates(t - d) - v)/tau_legs + sigma*(2./tau_legs)**.5*xi:1
d : second
"""
legs = NeuronGroup(8, model=eqs_legs, threshold='v > 1', reset='v = 0',
                   refractory=1*ms, method='euler')
legs.d = delay
spikes_legs = SpikeMonitor(legs)

# Command neurons
tau = 1 * ms
taus = 1.001 * ms
wex = 7
winh = -2
eqs_neuron = '''
dv/dt = (x - v)/tau : 1
dx/dt = (y - x)/taus : 1 # alpha currents
dy/dt = -y/taus : 1
'''
neurons = NeuronGroup(8, model=eqs_neuron, threshold='v>1', reset='v=0',
                      method='exact')
synapses_ex = Synapses(legs, neurons, on_pre='y+=wex')
synapses_ex.connect(j='i')
synapses_inh = Synapses(legs, neurons, on_pre='y+=winh', delay=deltaI)
synapses_inh.connect('abs(((j - i) % N_post) - N_post/2) <= 1')
spikes = SpikeMonitor(neurons)

run(duration, report='text')

nspikes = spikes.count
phi_est = imag(log(sum(nspikes * exp(gamma * 1j))))
print("True angle (deg): %.2f" % (phi/degree))
print("Estimated angle (deg): %.2f" % (phi_est/degree))
rmax = amax(nspikes)/duration/Hz
polar(concatenate((gamma, [gamma[0] + 2 * pi])),
      concatenate((nspikes, [nspikes[0]])) / duration / Hz,
      c='k')
axvline(phi, ls='-', c='g')
axvline(phi_est, ls='-', c='b')
show()



"""
A dummy package to allow wildcard import from brian2 without also importing
the pylab (numpy + matplotlib) namespace.

Usage: ``from brian2.only import *``

"""

# To minimize the problems with imports, import the packages in a sensible
# order

# isort:skip_file
# flake8: noqa

# The units and utils package does not depend on any other Brian package and
# should be imported first

from brian2.units import *
from brian2.utils import *
from brian2.core.tracking import *
from brian2.core.names import *
from brian2.core.spikesource import *

# The following packages only depend on something in the above set
from brian2.core.variables import linked_var
from brian2.core.functions import *
from brian2.core.preferences import *
from brian2.core.clocks import *
from brian2.equations import *

# The base class only depends on the above sets
from brian2.core.base import *

# The rest...
from brian2.core.network import *
from brian2.core.magic import *
from brian2.core.operations import *
from brian2.stateupdaters import *
from brian2.codegen import *
from brian2.core.namespace import *
from brian2.groups import *
from brian2.groups.subgroup import *
from brian2.synapses import *
from brian2.monitors import *
from brian2.importexport import *
from brian2.input import *
from brian2.spatialneuron import *
from brian2.devices import set_device, get_device, device, all_devices, seed
import brian2.devices.cpp_standalone as _cpp_standalone

# preferences
import brian2.core.core_preferences as _core_preferences

prefs.load_preferences()
prefs.do_validation()

prefs._backup()

set_device(all_devices["runtime"])


def restore_initial_state():
    """
    Restores internal Brian variables to the state they are in when Brian is imported

    Resets ``defaultclock.dt = 0.1*ms``,
    `BrianGlobalPreferences._restore` preferences, and set
    `BrianObject._scope_current_key` back to 0.
    """
    import gc

    prefs._restore()
    BrianObject._scope_current_key = 0
    defaultclock.dt = 0.1 * ms
    gc.collect()


# make the test suite available via brian2.test()
from brian2.tests import run as test

from brian2.units import __all__ as _all_units

__all__ = [
    "get_logger",
    "BrianLogger",
    "std_silent",
    "Trackable",
    "Nameable",
    "SpikeSource",
    "linked_var",
    "DEFAULT_FUNCTIONS",
    "Function",
    "implementation",
    "declare_types",
    "PreferenceError",
    "BrianPreference",
    "prefs",
    "brian_prefs",
    "Clock",
    "defaultclock",
    "Equations",
    "Expression",
    "Statements",
    "BrianObject",
    "BrianObjectException",
    "Network",
    "profiling_summary",
    "scheduling_summary",
    "MagicNetwork",
    "magic_network",
    "MagicError",
    "run",
    "stop",
    "collect",
    "store",
    "restore",
    "start_scope",
    "NetworkOperation",
    "network_operation",
    "StateUpdateMethod",
    "linear",
    "exact",
    "independent",
    "milstein",
    "heun",
    "euler",
    "rk2",
    "rk4",
    "ExplicitStateUpdater",
    "exponential_euler",
    "gsl_rk2",
    "gsl_rk4",
    "gsl_rkf45",
    "gsl_rkck",
    "gsl_rk8pd",
    "NumpyCodeObject",
    "CythonCodeObject",
    "get_local_namespace",
    "DEFAULT_FUNCTIONS",
    "DEFAULT_UNITS",
    "DEFAULT_CONSTANTS",
    "CodeRunner",
    "Group",
    "VariableOwner",
    "NeuronGroup",
    "Subgroup",
    "Synapses",
    "SpikeMonitor",
    "EventMonitor",
    "StateMonitor",
    "PopulationRateMonitor",
    "ImportExport",
    "BinomialFunction",
    "PoissonGroup",
    "PoissonInput",
    "SpikeGeneratorGroup",
    "TimedArray",
    "Morphology",
    "Soma",
    "Cylinder",
    "Section",
    "SpatialNeuron",
    "set_device",
    "get_device",
    "device",
    "all_devices",
    "seed",
    "restore_initial_state",
    "test",
]
__all__.extend(_all_units)



import pytest
from numpy.testing import assert_equal
from packaging.version import parse as parse_version

from brian2 import *
from brian2.devices.device import reinit_and_delete
from brian2.tests.utils import assert_allclose


class DerivedBrianObject(BrianObject):
    def __init__(self, name="derivedbrianobject*"):
        super().__init__(name=name)

    def __str__(self):
        return self.name

    __repr__ = __str__


@pytest.mark.codegen_independent
def test_base():
    x = DerivedBrianObject("x")
    y = DerivedBrianObject("y")
    assert_equal(x.when, "start")
    assert_equal(x.order, 0)
    assert_equal(len(x.contained_objects), 0)
    with pytest.raises(AttributeError):
        setattr(x, "contained_objects", [])
    x.contained_objects.append(y)
    assert_equal(len(x.contained_objects), 1)
    assert x.contained_objects[0] is y

    assert_equal(x.active, True)
    assert_equal(y.active, True)
    y.active = False
    assert_equal(x.active, True)
    assert_equal(y.active, False)
    y.active = True
    assert_equal(x.active, True)
    assert_equal(y.active, True)
    x.active = False
    assert_equal(x.active, False)
    assert_equal(y.active, False)


@pytest.mark.codegen_independent
def test_names():
    obj = BrianObject()
    obj2 = BrianObject()
    obj3 = DerivedBrianObject()
    assert_equal(obj.name, "brianobject")
    assert_equal(obj2.name, "brianobject_1")
    assert_equal(obj3.name, "derivedbrianobject")


@pytest.mark.codegen_independent
def test_duplicate_names():
    # duplicate names are allowed, as long as they are not part of the
    # same network
    obj1 = BrianObject(name="name1")
    obj2 = BrianObject(name="name2")
    obj3 = BrianObject(name="name")
    obj4 = BrianObject(name="name")
    net = Network(obj1, obj2)
    # all is good
    net.run(0 * ms)
    net = Network(obj3, obj4)
    with pytest.raises(ValueError):
        net.run(0 * ms)


@pytest.mark.standalone_compatible
@pytest.mark.multiple_runs
def test_active_flag():
    G = NeuronGroup(1, "dv/dt = 1/ms : 1")
    mon = StateMonitor(G, "v", record=0)
    mon.active = False
    run(1 * ms)
    mon.active = True
    G.active = False
    run(1 * ms)
    device.build(direct_call=False, **device.build_options)
    # Monitor should start recording at 1ms
    # Neurongroup should not integrate after 1ms (but should have integrated before)
    assert_allclose(mon[0].t[0], 1 * ms)
    assert_allclose(mon[0].v, 1.0)


@pytest.mark.codegen_independent
def test_version():
    # Check that determining the Brian version works correctly
    import brian2

    version = brian2.__version__
    assert version.startswith("2.")

    # Check that the version tuple is correct
    version_tuple = brian2.__version_tuple__

    expected = []
    for v in version.split("."):
        try:
            expected.append(int(v))
        except ValueError:
            expected.append(v)

    # Note that setuptools_scm >= v8.2.1 include a post suffix in the
    # version tutple, but earlier versions don't
    assert version_tuple == tuple(expected) or version_tuple == tuple(
        t for t in expected if isinstance(t, int)
    )


if __name__ == "__main__":
    test_base()
    test_names()
    test_duplicate_names()
    test_active_flag()



{% macro cython_directives() %}
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=False
#cython: cpow=True
#cython: infer_types=True
{% endmacro %}

{% macro imports() %}
import numpy as _numpy
cimport numpy as _numpy
from libc.math cimport fabs, sin, cos, tan, sinh, cosh, tanh, exp, log, log10, expm1, log1p, sqrt, asin, acos, atan, fmod, floor, ceil, isinf
cdef extern from "math.h":
    double M_PI
# Import the two versions of std::abs
from libc.stdlib cimport abs  # For integers
from libc.math cimport abs  # For floating point values
from libc.limits cimport INT_MIN, INT_MAX
from libcpp cimport bool
from libcpp.set cimport set
from cython.operator cimport dereference as _deref, preincrement as _preinc
cimport cython as _cython

_numpy.import_array()
cdef extern from "numpy/ndarraytypes.h":
    void PyArray_CLEARFLAGS(_numpy.PyArrayObject *arr, int flags)
from libc.stdlib cimport free

cdef extern from "numpy/npy_math.h":
    bint npy_isinf(double x)
    double NPY_INFINITY

cdef extern from "stdint_compat.h":
    # Longness only used for type promotion
    # Actual compile time size used for conversion
    ctypedef signed int int32_t
    ctypedef signed long int64_t
    ctypedef unsigned long uint64_t
    # It seems we cannot used a fused type here
    cdef int int_(bool)
    cdef int int_(char)
    cdef int int_(short)
    cdef int int_(int)
    cdef int int_(long)
    cdef int int_(float)
    cdef int int_(double)
    cdef int int_(long double)
{% endmacro %}

{% macro before_run() %}
{{ cython_directives() }}
{{ imports() }}

# support code
{{ support_code_lines | autoindent }}

def main(_namespace):
    {{ load_namespace | autoindent }}
    if '_owner' in _namespace:
        _owner = _namespace['_owner']
    {% block before_code %}
    # EMPTY_CODE_BLOCK  -- overwrite in child template
    {% endblock %}
{% endmacro %}

{% macro run() %}
{{ cython_directives() }}
{{ imports() }}

# support code
{{ support_code_lines | autoindent }}

# template-specific support code
{% block template_support_code %}
{% endblock %}

def main(_namespace):
    cdef size_t _idx
    cdef size_t _vectorisation_idx
    {{ load_namespace | autoindent }}
    if '_owner' in _namespace:
        _owner = _namespace['_owner']
    {% block maincode %}
    {% endblock %}

{% endmacro %}

{% macro after_run() %}
{{ cython_directives() }}
{{ imports() }}

# support code
{{support_code_lines | autoindent}}

def main(_namespace):
    {{ load_namespace | autoindent }}
    if '_owner' in _namespace:
        _owner = _namespace['_owner']
    {% block after_code %}
    # EMPTY_CODE_BLOCK  -- overwrite in child template
    {% endblock %}
{% endmacro %}



"""
Base package for all monitors, i.e. objects to record activity during a
simulation run.
"""

from .ratemonitor import *
from .spikemonitor import *
from .statemonitor import *

__all__ = ["SpikeMonitor", "EventMonitor", "StateMonitor", "PopulationRateMonitor"]



'''
Some simple benchmarks to check how much the performance is impaired when using units (units are not used in the main
simulation loop but also analysing/plotting should not be noticeably slow for the user)
'''
import time

import numpy as np

from brian2 import *

sizes = [10, 100, 1000, 10000, 100000]

access = lambda idx, ar1, ar2: ar1[idx]
multiply = lambda idx, ar1, ar2: 42 * ar1
addition = lambda idx, ar1, ar2: ar1 + ar2
slicing = lambda idx, ar1, ar2: ar1[idx:]

for func_name, func in [('access', access),
                        ('multiply', multiply),
                        ('add', addition),
                        ('slicing', slicing)]:
    times_no_unit = []
    times_unit = []
    for size in sizes:
        no_unit1 = np.random.randn(size)
        no_unit2 = np.random.randn(size)
        with_unit1 = no_unit1 * mV
        with_unit2 = no_unit2 * mV
        start = time.time()
        for x in range(size):
           func(x, no_unit1, no_unit2)
        times_no_unit.append(time.time() - start)
        start = time.time()
        for x in range(size):
            func(x, with_unit1, with_unit2)
        times_unit.append(time.time() - start)
    print('')
    print(func_name, ':')
    print('No unit ', times_no_unit)
    print('   unit ', times_unit)
    print('relative', np.array(times_unit) / np.array(times_no_unit))

# Dimensionless Quantities
print('*' * 60)
print('Dimensionless quantities')

for func_name, func in [('access', access),
                        ('multiply', multiply),
                        ('add', addition),
                        ('slicing', slicing)]:
    times_no_unit = []
    times_unit = []
    for size in sizes:
        no_unit1 = np.random.randn(size)
        no_unit2 = np.random.randn(size)
        with_unit1 = no_unit1 * mV/mV
        with_unit2 = no_unit2 * mV/mV
        start = time.time()
        for x in range(size):
           func(x, no_unit1, no_unit2)
        times_no_unit.append(time.time() - start)
        start = time.time()
        for x in range(size):
            func(x, with_unit1, with_unit2)
        times_unit.append(time.time() - start)
    print('')
    print(func_name, ':')
    print('No unit ', times_no_unit)
    print('   unit ', times_unit)
    print('relative', np.array(times_unit) / np.array(times_no_unit))


#!/usr/bin/env python
"""
NMDA synapses.
"""
from brian2 import *

a = 1 / (10*ms)
b = 1 / (10*ms)
c = 1 / (10*ms)

neuron_input = NeuronGroup(2, 'dv/dt = 1/(10*ms) : 1', threshold='v>1', reset='v = 0',
                    method='euler')
neurons = NeuronGroup(1, """dv/dt = (g-v)/(10*ms) : 1
                            g : 1""", method='exact')
S = Synapses(neuron_input, neurons, '''
                dg_syn/dt = -a*g_syn+b*x*(1-g_syn) : 1 (clock-driven)
                g_post = g_syn : 1 (summed)
                dx/dt=-c*x : 1 (clock-driven)
                w : 1 # synaptic weight
             ''', on_pre='x += w') # NMDA synapses

S.connect()
S.w = [1., 10.]
neuron_input.v = [0., 0.5]

M = StateMonitor(S, 'g',
                 # If not using standalone mode, this could also simply be
                 # record=True
                 record=np.arange(len(neuron_input)*len(neurons)))
Mn = StateMonitor(neurons, 'g', record=0)

run(1000*ms)

subplot(2, 1, 1)
plot(M.t/ms, M.g.T)
xlabel('Time (ms)')
ylabel('g_syn')
subplot(2, 1, 2)
plot(Mn.t/ms, Mn[0].g)
ylabel('Time (ms)')
ylabel('g')
tight_layout()
show()



import builtins

import pytest

from brian2.utils.environment import running_from_ipython
from brian2.utils.stringtools import SpellChecker


@pytest.mark.codegen_independent
def test_environment():
    """
    Test information about the environment we are running under.
    """
    if hasattr(builtins, "__IPYTHON__"):
        testing_under_ipython = True
        del builtins.__IPYTHON__
    else:
        testing_under_ipython = False

    assert not running_from_ipython()
    builtins.__IPYTHON__ = True
    assert running_from_ipython()

    if not testing_under_ipython:
        del builtins.__IPYTHON__


@pytest.mark.codegen_independent
def test_spell_check():
    checker = SpellChecker(["vm", "alpha", "beta"])
    assert checker.suggest("Vm") == {"vm"}
    assert checker.suggest("alphas") == {"alpha"}
    assert checker.suggest("bta") == {"beta"}
    assert checker.suggest("gamma") == set()


if __name__ == "__main__":
    test_environment()
    test_spell_check()
